# SPDX-License-Identifier: LGPL-2.1-or-later

#***************************************************************************
#*   Copyright (c) 2012 Sebastian Hoogen <github@sebastianhoogen.de>       *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

__title__ = "FreeCAD OpenSCAD Workbench - 2D helper functions"
__author__ = "Sebastian Hoogen"
__url__ = ["https://www.freecad.org"]

'''
This Script includes python functions to convert imported dxf geometry to Faces
'''

from functools import reduce

class Overlappingfaces():
    '''combines overlapping faces together'''
    def __init__(self,facelist):
        self.sortedfaces = sorted(facelist,key=(lambda shape: shape.Area),reverse=True)
        self.builddepdict()

    @staticmethod
    def dofacesoverlapboundbox(bigface,smallface):
        return bigface.BoundBox.isIntersection(smallface.BoundBox)

    @staticmethod
    def dofacesoverlapallverts(bigface,smallface):
        def vertsinface(f1,verts,tol=0.001,inface=True):
            '''check if all given verts are inside shape f1'''
            return all([f1.isInside(vert.Point,tol,inface) for vert in verts])
        return vertsinface(bigface,smallface.Vertexes)

    @staticmethod
    def dofacesoverlapproximity(bigface,smallface):
        l1,l2 = bigface.proximity(smallface)
        return len(l1) > 0 or len(l2) > 0

    @staticmethod
    def dofacesoverlapboolean(bigface,smallface):
        return bigface.common(smallface).Area > 0

    def builddepdict(self):
        import Part
        import itertools
        self.isinsidedict = {}
        #for bigface, smallface in itertools.combinations(sortedfaces,2):
        for bigfacei, smallfacei in\
                itertools.combinations(range(len(self.sortedfaces)),2):
            try:
                overlap = Overlappingfaces.dofacesoverlapproximity(\
                        self.sortedfaces[bigfacei],self.sortedfaces[smallfacei])
            except (NotImplementedError, Part.OCCError) as e:
                try:
                    overlap = Overlappingfaces.dofacesoverlapboolean(\
                            self.sortedfaces[bigfacei],\
                            self.sortedfaces[smallfacei])
                except Part.OCCError:
                    overlap = Overlappingfaces.dofacesoverlapallverts(\
                            self.sortedfaces[bigfacei],\
                            self.sortedfaces[smallfacei])
            if overlap:
                smallinbig = self.isinsidedict.get(bigfacei,[])
                smallinbig.append(smallfacei)
                if len(smallinbig) == 1:
                    self.isinsidedict[bigfacei] = smallinbig

    @staticmethod
    def finddepth(dict1,faceidx,curdepth=0):
        if faceidx not in dict1:
            return curdepth+1
        else:
            return max([(Overlappingfaces.finddepth(dict1,childface,curdepth+1)) for childface in dict1[faceidx]])

    def findrootdepth(self):
        return max([Overlappingfaces.finddepth(self.isinsidedict,fi) for fi in range(len(self.sortedfaces))])

    def hasnoparent(self,faceindex):
        return Overlappingfaces.hasnoparentstatic(self.isinsidedict,faceindex)

    @staticmethod
    def hasnoparentstatic(isinsidedict,faceindex):
        for smalllist in isinsidedict.values():
            if faceindex in smalllist:
                return False
        return True


    @staticmethod
    def directchildren(isinsidedict,parent):
        #return [child for child in isinsidedict.get(parent,[]) if child not in isinsidedict]
        dchildren=[]
        for child in isinsidedict.get(parent,[]):
            direct = True
            for key, value in isinsidedict.items():
                if key != parent and child in value and parent not in value:
                    direct = False
            if direct:
                dchildren.append(child)
        return dchildren


    @staticmethod
    def printtree(isinsidedict,facenum):
        def printtreechild(isinsidedict,facenum,parent):
            children=Overlappingfaces.directchildren(isinsidedict,parent)
            print('parent %d directchild %s' % (parent,children))
            if children:
                subdict=isinsidedict.copy()
                del subdict[parent]
                for child in children:
                    printtreechild(subdict,facenum,child)

        rootitems=[fi for fi in range(facenum) if Overlappingfaces.hasnoparentstatic(isinsidedict,fi)]
        for rootitem in rootitems:
            printtreechild(isinsidedict,facenum,rootitem)

    def makefeatures(self,doc):
        import FreeCAD
        def addshape(faceindex):
            obj=doc.addObject('Part::Feature','facefromedges_%d' % faceindex)
            obj.Shape = self.sortedfaces[faceindex]
            obj.ViewObject.hide()
            return obj

        def addfeature(faceindex,isinsidedict):
            directchildren = Overlappingfaces.directchildren(isinsidedict,faceindex)
            if len(directchildren) == 0:
                obj=addshape(faceindex)
            else:
                subdict=isinsidedict.copy()
                del subdict[faceindex]
                obj=doc.addObject("Part::Cut","facesfromedges_%d" % faceindex)
                obj.Base= addshape(faceindex) #we only do subtraction
                if len(directchildren) == 1:
                    obj.Tool = addfeature(directchildren[0],subdict)
                else:
                    obj.Tool = doc.addObject("Part::MultiFuse",\
                            "facesfromedges_union")
                    obj.Tool.Shapes = [addfeature(child,subdict)\
                            for child in directchildren]
                    obj.Tool.ViewObject.hide()
            obj.ViewObject.hide()
            return obj

        rootitems = [fi for fi in range(len(self.sortedfaces)) if self.hasnoparent(fi)]
        for rootitem in rootitems:
            addfeature(rootitem,self.isinsidedict).ViewObject.show()


    def makeshape(self):
        def removefaces(rfaces):
            for tfi in directchildren[::-1]:
                finishedwith.append(tfi)
                if tfi in isinsidedict:
                    del isinsidedict[tfi]
                for key,value in isinsidedict.items():
                    if tfi in value:
                        newlist=value[:] #we work on a shallow copy of isinsidedict
                        newlist.remove(tfi)
                        isinsidedict[key]=newlist

        def hasnoparent(faceindex):
            for smalllist in self.isinsidedict.values():
                if faceindex in smalllist:
                    return False
            return True

        faces=self.sortedfaces[:]
        isinsidedict=self.isinsidedict.copy()
        finishedwith=[]
        while not all([Overlappingfaces.hasnoparentstatic(isinsidedict,fi) for fi in range(len(faces))]):
            for fi in range(len(faces))[::-1]:
                directchildren = Overlappingfaces.directchildren(isinsidedict,fi)
                if not directchildren:
                    continue
                elif len(directchildren) == 1:
                    faces[fi]=faces[fi].cut(faces[directchildren[0]])
                    removefaces(directchildren)
                else:
                    toolface=fusefaces([faces[tfi] for tfi in directchildren])
                    faces[fi]=faces[fi].cut(toolface)
                    removefaces(directchildren)

        faces =[face for index,face in enumerate(faces) if index not in finishedwith]
        return fusefaces(faces)

def findConnectedEdges(edgelist,eps=1e-6,debug=False):
    '''returns a list of list of connected edges'''

    def vertequals(v1,v2,eps=1e-6):
        '''check two vertices for equality'''
        return v1.Point.sub(v2.Point).Length<eps

    def vertindex(forward):
        '''return index of last or first element'''
        return -1 if forward else 0

    freeedges = edgelist[:]
    retlist = []
    debuglist = []
    while freeedges:
        startwire = freeedges.pop(0)
        forward = True
        newedge = [(startwire,True)]
        for forward in (True, False):
            found = True
            while found:
                lastvert = newedge[vertindex(forward)][0].Vertexes[vertindex(forward == newedge[vertindex(forward)][1])]
                for ceindex, checkedge in enumerate(freeedges):
                    found = False
                    for cvindex, cvert in enumerate([checkedge.Vertexes[0],checkedge.Vertexes[-1]]):
                        if vertequals(lastvert,cvert,eps):
                            if forward:
                                newedge.append((checkedge,cvindex == 0))
                            else:
                                newedge.insert(0,(checkedge,cvindex == 1))
                            del freeedges[ceindex]
                            found = True
                            break
                    else:
                        found = False
                    if found:
                        break
                else:
                    found = False
        #we are finished for this edge
        debuglist.append(newedge)
        retlist.append([item[0] for item in newedge]) #strip off direction
    if debug:
        return retlist,debuglist
    else:
        return retlist

def endpointdistance(edges):
    '''return the distance of vertices in path (list of edges) as
    maximum, minimum and distance between start and endpoint
    it expects the edges to be traversed forward starting from Vertex 0'''
    numedges=len(edges)
    if numedges == 1 and len(edges[0].Vertexes) == 1:
        return 0.0,0.0,0.0
    outerdistance = edges[0].Vertexes[0].Point.sub(\
        edges[-1].Vertexes[-1].Point).Length
    if numedges > 1:
        innerdistances=[edges[i].Vertexes[-1].Point.sub(edges[i+1].\
                Vertexes[0].Point).Length for i in range(numedges-1)]
        return max(innerdistances),min(innerdistances),outerdistance
    else:
        return 0.0,0.0,outerdistance

def endpointdistancedebuglist(debuglist):
    '''return the distance of vertices in path (list of edges) as
    maximum, minimum and distance between start and endpoint
    it expects a 'not reversed' flag for every edge'''
    numedges=len(debuglist)
    if numedges == 1 and len(debuglist[0][0].Vertexes) == 1:
        return 0.0,0.0,0.0
    outerdistance = debuglist[0][0].Vertexes[(not debuglist[0][1])*-1].\
            Point.sub(debuglist[-1][0].Vertexes[(debuglist[-1][1])*-1].\
            Point).Length
    if numedges > 1:
        innerdistances=[debuglist[i][0].Vertexes[debuglist[i][1]*-1].\
                Point.sub(debuglist[i+1][0].Vertexes[(not debuglist[i+1][1])*\
                -1].Point).Length for i in range(numedges-1)]
        return max(innerdistances),min(innerdistances),outerdistance
    else:
        return 0.0,0.0,outerdistance

def edgestowires(edgelist,eps=0.001):
    '''takes list of edges and returns a list of wires'''
    import Part
    import Draft
    # TODO remove double edges
    wirelist=[]
    #for path in findConnectedEdges(edgelist,eps=eps):
    for path,debug in zip(*findConnectedEdges(edgelist,eps=eps,debug=True)):
        maxd,mind,outerd = endpointdistancedebuglist(debug)
        assert(maxd <= eps*2) # Assume the input to be broken
        if maxd < eps*2 and maxd > 0.000001: # OCC won't like it if maxd > 0.02:
            print('endpointdistance max:%f min:%f, ends:%f' %(maxd,mind,outerd))

            if True:
                tobeclosed = outerd < eps*2
                # OpenSCAD uses 0.001 for corase grid
                #from draftlibs import fcvec, fcgeo
                w2=superWireReverse(debug,tobeclosed)
                wirelist.append(w2)
            else:#this locks up FreeCAD
                comp=Part.Compound(path)
                wirelist.append(comp.connectEdgesToWires(False,eps).Wires[0])
        else:
            done = False
            try:
                wire=Part.Wire(path)
                #if not close or wire.isClosed or outerd > 0.0001:
                wirelist.append(Part.Wire(path))
                done = True
            except Part.OCCError:
                pass
            if not done:
                comp=Part.Compound(path)
                wirelist.append(comp.connectEdgesToWires(False,eps).Wires[0])
    return wirelist

def subtractfaces(faces):
    '''searches for the biggest face and subtracts all smaller ones from the
    first. Only makes sense if all faces overlap.'''
    if len(faces)==1:
        return faces[0]
    else:
        facelist=sorted(faces,key=(lambda shape: shape.Area),reverse=True)
        base=facelist[0]
        tool=reduce(lambda p1,p2: p1.fuse(p2),facelist[1:])
        return base.cut(tool)

def fusefaces(faces):
    if len(faces)==1:
        return faces[0]
    else:
        return reduce(lambda p1,p2: p1.fuse(p2),faces)

def subtractfaces2(faces):
    '''Sort faces, check if they overlap. Subtract overlapping face and fuse
    nonoverlapping groups.'''
    return fusefaces([subtractfaces(facegroup) for facegroup in findoverlappingfaces(faces)])

def edgestofaces(edges,algo=3,eps=0.001):
    import Part
    wires = edgestowires(edges,eps)
    facel=[]
    for w in wires:
        if not w.isClosed():
            p0 = w.Vertexes[0].Point
            p1 = w.Vertexes[-1].Point
            edges2 = w.Edges[:]
            try:
                edges2.append(Part.LineSegment(p1,p0).toShape())
                w = Part.Wire(edges2)
            except OCCError:
                comp=Part.Compound(edges2)
                w = comp.connectEdgesToWires(False,eps).Wires[0]
        facel.append(Part.Face(w))

    if algo is None:
        return facel
    elif algo == 1: #stable behavior
        return subtractfaces(facel)
    elif algo == 0: #return all faces
        return Part.Compound(facel)
    elif algo == 2:
        return subtractfaces2(facel)
    elif algo == 3:
        return Overlappingfaces(facel).makeshape()

def superWireReverse(debuglist,closed=False):
    '''superWireReverse(debuglist,[closed]): forces a wire between edges
    that don't necessarily have coincident endpoints. If closed=True, wire
    will always be closed. debuglist has a tuple for every edge.The first
    entry is the edge, the second is the flag 'does not need to be inverted'
    '''
    #taken from draftlibs
    def median(v1,v2):
        vd = v2.sub(v1)
        vd.scale(.5,.5,.5)
        return v1.add(vd)
    try:
        from DraftGeomUtils import findMidpoint
    except ImportError: #workaround for Version 0.12
        from draftlibs.fcgeo import findMidpoint #workaround for Version 0.12
    import Part
    print(debuglist)
    newedges = []
    for i in range(len(debuglist)):
        curr = debuglist[i]
        if i == 0:
            if closed:
                prev = debuglist[-1]
            else:
                prev = None
        else:
            prev = debuglist[i-1]
        if i == (len(debuglist)-1):
            if closed:
                nexte = debuglist[0]
            else:
                nexte = None
        else:
            nexte = debuglist[i+1]
        print(i,prev,curr,nexte)
        if prev:
            if curr[0].Vertexes[-1*(not curr[1])].Point == \
                    prev[0].Vertexes[-1*prev[1]].Point:
                p1 = curr[0].Vertexes[-1*(not curr[1])].Point
            else:
                p1 = median(curr[0].Vertexes[-1*(not curr[1])].Point,\
                        prev[0].Vertexes[-1*prev[1]].Point)
        else:
            p1 = curr[0].Vertexes[-1*(not curr[1])].Point
        if nexte:
            if curr[0].Vertexes[-1*curr[1]].Point == \
                nexte[0].Vertexes[-1*(not nexte[1])].Point:
                p2 = nexte[0].Vertexes[-1*(not nexte[1])].Point
            else:
                p2 = median(curr[0].Vertexes[-1*(curr[1])].Point,\
                        nexte[0].Vertexes[-1*(not nexte[1])].Point)
        else:
            p2 = curr[0].Vertexes[-1*(curr[1])].Point
        if isinstance(curr[0].Curve,(Part.LineSegment, Part.Line)):
            print("line",p1,p2)
            newedges.append(Part.LineSegment(p1,p2).toShape())
        elif isinstance(curr[0].Curve,Part.Circle):
            p3 = findMidpoint(curr[0])
            print("arc",p1,p3,p2)
            newedges.append(Part.Arc(p1,p3,p2).toShape())
        else:
            print("Cannot superWire edges that are not lines or arcs")
            return None
    print(newedges)
    return Part.Wire(newedges)

def importDXFface(filename,layer=None,doc=None):
    import FreeCAD
    import importDXF
    importDXF.readPreferences()
    importDXF.getDXFlibs()
    importDXF.dxfMakeBlocks = False
    doc = doc or FreeCAD.activeDocument()
    layers = importDXF.processdxf(doc,filename,False,False) or importDXF.layers
    for l in layers:
        if FreeCAD.GuiUp:
            for o in l.Group:
                o.ViewObject.hide()
            l.ViewObject.hide()
    groupobj=[go for go in layers if (not layer) or go.Label == layer]
    edges=[]
    if not groupobj:
        raise ValueError('import of layer %s failed' % layer)
    for shapeobj in groupobj[0].Group:
        edges.extend(shapeobj.Shape.Edges)
    faces = edgestofaces(edges)
    # in order to allow multiple import with the same layer name
    # we need to remove used objects from the layer group
    container = None
    for layer in layers: #remove everything that has been imported
        if container is None:
            container = layer.getParentGroup()
        removeOp = getattr(layer, "removeObjectsFromDocument", None)
        if callable(removeOp):
            layer.removeObjectsFromDocument()
        for obj in layer.Group:
            obj.Document.removeObject(obj.Name)
        layer.Document.removeObject(layer.Name)
    container.Document.removeObject(container.Name)
    return faces



import FreeCAD
import unittest
from .TechDrawTestUtilities import createPageWithSVGTemplate
from PySide import QtCore

class DrawViewDimensionTest(unittest.TestCase):
    def setUp(self):
        """Creates a page and 2 views"""
        print("DVDTest.setUp()")
        FreeCAD.newDocument("TDDimTest")
        FreeCAD.setActiveDocument("TDDimTest")
        FreeCAD.ActiveDocument = FreeCAD.getDocument("TDDimTest")
        self.document = FreeCAD.ActiveDocument

        # make source feature
        self.document.addObject("Part::Box", "Box")
        self.document.addObject("Part::Sphere", "Sphere")

        # make a page
        self.page = createPageWithSVGTemplate(self.document)
        self.page.Scale = 5.0

        # make Views
        self.view = self.document.addObject("TechDraw::DrawViewPart", "View")
        self.page.addView(self.view)
        self.view.Source = [self.document.Box]
        self.view.X = 30
        self.view.Y = 150

        self.view1 = self.document.addObject("TechDraw::DrawViewPart", "View001")
        self.page.addView(self.view1)
        self.view1.Source = [self.document.Sphere]
        self.view1.X = 220
        self.view1.Y = 150

        self.document.recompute()

        #wait for threads to complete before checking result
        loop = QtCore.QEventLoop()

        timer = QtCore.QTimer()
        timer.setSingleShot(True)
        timer.timeout.connect(loop.quit)

        timer.start(5000)   #5 second delay
        loop.exec_()

    def tearDown(self):
        print("DVDTest.tearDown()")
        FreeCAD.closeDocument("TDDimTest")

    def testLengthDimension(self):
        """Tests if a length dimension can be added to view"""
        # make length dimension
        print("making length dimension")

        dimension = self.document.addObject("TechDraw::DrawViewDimension", "Dimension")
        self.page.addView(dimension)
        dimension.Type = "Distance"
        dimension.References2D = [(self.view, "Edge1")]
        print("finished length dimension")
        self.document.recompute()
        self.assertTrue("Up-to-date" in dimension.State)

    def testRadiusDimension(self):
        """Tests if a radius dimension can be added to view1"""
        print("making radius dimension")
        dimension = self.document.addObject("TechDraw::DrawViewDimension", "Dimension001")
        self.page.addView(dimension)
        dimension.Type = "Radius"
        dimension.MeasureType = "Projected"
        dimension.References2D = [(self.view1, "Edge0")]
        self.document.recompute()
        self.assertTrue("Up-to-date" in dimension.State)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 Pekka Roivainen <pekkaroi@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PathScripts import PathUtils
from PySide.QtCore import QT_TRANSLATE_NOOP
import copy
import FreeCAD
import Path
import Path.Dressup.Utils as PathDressup
import math


# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")

if FreeCAD.GuiUp:
    import FreeCADGui


translate = FreeCAD.Qt.translate


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class AnnotatedGCode:
    def __init__(self, command, start_point):
        self.start_point = start_point
        self.command = command
        self.end_point = (
            command.Parameters.get("X", start_point[0]),
            command.Parameters.get("Y", start_point[1]),
            command.Parameters.get("Z", start_point[2]),
        )
        self.is_line = command.Name in Path.Geom.CmdMoveStraight
        self.is_arc = command.Name in Path.Geom.CmdMoveArc
        self.xy_length = None
        if self.is_line:
            self.xy_length = (
                (start_point[0] - self.end_point[0]) ** 2
                + (start_point[1] - self.end_point[1]) ** 2
            ) ** 0.5
        elif self.is_arc:
            self.center_xy = (
                start_point[0] + command.Parameters.get("I", 0),
                start_point[1] + command.Parameters.get("J", 0),
            )
            self.start_angle = math.atan2(
                start_point[1] - self.center_xy[1],
                start_point[0] - self.center_xy[0],
            )
            self.end_angle = math.atan2(
                self.end_point[1] - self.center_xy[1],
                self.end_point[0] - self.center_xy[0],
            )
            if self.command.Name in Path.Geom.CmdMoveCCW and self.end_angle < self.start_angle:
                self.end_angle += 2 * math.pi
            if self.command.Name in Path.Geom.CmdMoveCW and self.end_angle > self.start_angle:
                self.end_angle -= 2 * math.pi
            self.radius = (
                (start_point[0] - self.center_xy[0]) ** 2
                + (start_point[1] - self.center_xy[1]) ** 2
            ) ** 0.5
            self.xy_length = self.radius * abs(self.end_angle - self.start_angle)

    """Makes a copy of this annotated gcode at the given z height"""

    def clone(self, z_start=None, z_end=None, reverse=False):
        z_start = z_start if z_start is not None else self.start_point[2]
        z_end = z_end if z_end is not None else self.end_point[2]

        other = copy.copy(self)
        otherParams = copy.copy(self.command.Parameters)
        otherCommandName = self.command.Name
        other.start_point = (self.start_point[0], self.start_point[1], z_start)
        other.end_point = (self.end_point[0], self.end_point[1], z_end)
        otherParams.update({"Z": z_end})
        if reverse:
            other.start_point, other.end_point = other.end_point, other.start_point
            otherParams.update(
                {"X": other.end_point[0], "Y": other.end_point[1], "Z": other.end_point[2]}
            )
            if other.is_arc:
                other.start_angle, other.end_angle = other.end_angle, other.start_angle
                otherCommandName = (
                    Path.Geom.CmdMoveCW[0]
                    if self.command.Name in Path.Geom.CmdMoveCCW
                    else Path.Geom.CmdMoveCCW[0]
                )
                otherParams.update(
                    {
                        "I": other.center_xy[0] - other.start_point[0],
                        "J": other.center_xy[1] - other.start_point[1],
                    }
                )
        other.command = Path.Command(otherCommandName, otherParams)
        return other

    """Splits the edge into two parts, the first split_length (if less than xy_length) long. Only supported for lines and arcs (no rapids)"""

    def split(self, split_length):
        split_length = min(split_length, self.xy_length)
        p = split_length / self.xy_length
        firstParams = copy.copy(self.command.Parameters)
        secondParams = copy.copy(self.command.Parameters)
        split_point = None
        if self.is_line:
            split_point = (
                self.start_point[0] * (1 - p) + self.end_point[0] * p,
                self.start_point[1] * (1 - p) + self.end_point[1] * p,
                self.start_point[2] * (1 - p) + self.end_point[2] * p,
            )
        elif self.is_arc:
            angle = self.start_angle * (1 - p) + self.end_angle * p
            split_point = (
                self.center_xy[0] + self.radius * math.cos(angle),
                self.center_xy[1] + self.radius * math.sin(angle),
                self.start_point[2] * (1 - p) + self.end_point[2] * p,
            )
            secondParams.update(
                {
                    "I": self.center_xy[0] - split_point[0],
                    "J": self.center_xy[1] - split_point[1],
                }
            )
        else:
            raise Exception("Invalid type, can only split (non-rapid) lines and arcs")

        firstParams.update({"X": split_point[0], "Y": split_point[1], "Z": split_point[2]})
        first_command = Path.Command(self.command.Name, firstParams)
        second_command = Path.Command(self.command.Name, secondParams)
        return AnnotatedGCode(first_command, self.start_point), AnnotatedGCode(
            second_command, split_point
        )


class ObjectDressup:
    def __init__(self, obj):
        self.obj = obj
        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The base toolpath to modify"),
        )
        obj.addProperty(
            "App::PropertyAngle",
            "Angle",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Angle of ramp"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "Method",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Ramping Method"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "RampFeedRate",
            "FeedRate",
            QT_TRANSLATE_NOOP("App::Property", "Which feed rate to use for ramping"),
        )
        obj.addProperty(
            "App::PropertySpeed",
            "CustomFeedRate",
            "FeedRate",
            QT_TRANSLATE_NOOP("App::Property", "Custom feed rate"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "UseStartDepth",
            "StartDepth",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Should the dressup ignore motion commands above DressupStartDepth",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "DressupStartDepth",
            "StartDepth",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The depth where the ramp dressup is enabled. Above this ramps are not generated, but motion commands are passed through as is.",
            ),
        )

        # populate the enumerations
        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

        obj.Proxy = self
        self.setEditorProperties(obj)

        # initialized later
        self.wire = None
        self.angle = None
        self.rapids = None
        self.method = None
        self.outedges = None
        self.ignoreAboveEnabled = None
        self.ignoreAbove = None

    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """PropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        enums = {
            "Method": [
                (translate("CAM_DressupRampEntry", "RampMethod1"), "RampMethod1"),
                (translate("CAM_DressupRampEntry", "RampMethod2"), "RampMethod2"),
                (translate("CAM_DressupRampEntry", "RampMethod3"), "RampMethod3"),
                (translate("CAM_DressupRampEntry", "Helix"), "Helix"),
            ],
            "RampFeedRate": [
                (
                    translate("CAM_DressupRampEntry", "Horizontal Feed Rate"),
                    "Horizontal Feed Rate",
                ),
                (
                    translate("CAM_DressupRampEntry", "Vertical Feed Rate"),
                    "Vertical Feed Rate",
                ),
                (
                    translate("CAM_DressupRampEntry", "Ramp Feed Rate"),
                    "Ramp Feed Rate",
                ),
                (translate("CAM_DressupRampEntry", "Custom"), "Custom"),
            ],
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, obj, prop):
        if prop in ["RampFeedRate", "UseStartDepth"]:
            self.setEditorProperties(obj)
        if prop == "Path" and obj.ViewObject:
            obj.ViewObject.signalChangeIcon()

    def setEditorProperties(self, obj):
        if hasattr(obj, "UseStartDepth"):
            if obj.UseStartDepth:
                obj.setEditorMode("DressupStartDepth", 0)
            else:
                obj.setEditorMode("DressupStartDepth", 2)

        if obj.RampFeedRate == "Custom":
            obj.setEditorMode("CustomFeedRate", 0)
        else:
            obj.setEditorMode("CustomFeedRate", 2)

    def onDocumentRestored(self, obj):
        self.setEditorProperties(obj)

    def setup(self, obj):
        obj.Angle = 60
        obj.Method = 2
        if PathDressup.baseOp(obj.Base).StartDepth is not None:
            obj.DressupStartDepth = PathDressup.baseOp(obj.Base).StartDepth

    def execute(self, obj):
        if not obj.Base:
            return
        if not obj.Base.isDerivedFrom("Path::Feature"):
            return
        if not obj.Base.Path:
            return

        if not PathDressup.baseOp(obj.Base).Active:
            path = Path.Path("(inactive operation)")
            obj.Path = path
            return

        if obj.Angle >= 90:
            obj.Angle = 89.9
        elif obj.Angle <= 0:
            obj.Angle = 0.1

        if hasattr(obj, "UseStartDepth"):
            self.ignoreAboveEnabled = obj.UseStartDepth
            self.ignoreAbove = obj.DressupStartDepth
        else:
            self.ignoreAboveEnabled = False
            self.ignoreAbove = 0

        self.angle = obj.Angle
        self.method = obj.Method
        positioned_path = PathUtils.getPathWithPlacement(obj.Base)
        cmds = positioned_path.Commands if hasattr(positioned_path, "Commands") else []
        self.edges = []
        start_point = (0, 0, 0)
        last_params = {}
        for cmd in cmds:
            # Skip repeat move commands
            params = cmd.Parameters
            if (
                cmd.Name in Path.Geom.CmdMoveAll
                and len(self.edges) > 0
                and cmd.Name == self.edges[-1].command.Name
            ):
                found_diff = False
                for k, v in params.items():
                    if last_params.get(k, None) != v:
                        found_diff = True
                        break
                if not found_diff:
                    continue

            last_params.update(params)

            if cmd.Name in Path.Geom.CmdMoveAll and (
                "X" not in params or "Y" not in params or "Z" not in params
            ):
                params["X"] = params.get("X", start_point[0])
                params["Y"] = params.get("Y", start_point[1])
                params["Z"] = params.get("Z", start_point[2])
                cmd = Path.Command(cmd.Name, params)
            annotated = AnnotatedGCode(cmd, start_point)
            self.edges.append(annotated)
            start_point = annotated.end_point
        if self.method in ["RampMethod1", "RampMethod2", "RampMethod3"]:
            self.outedges = self.generateRamps()
        else:
            self.outedges = self.generateHelix()
        obj.Path = self.createCommands(obj, self.outedges)

    def generateRamps(self):
        edges = self.edges
        outedges = []
        for edgei, edge in enumerate(edges):
            if edge.is_line or edge.is_arc:
                rampangle = self.angle
                # check for plunge
                if edge.xy_length < 1e-6 and edge.end_point[2] < edge.start_point[2]:
                    # check if above ignoreAbove parameter - do not generate ramp if it is
                    noramp_edge, edge = self.processIgnoreAbove(edge)
                    if noramp_edge is not None:
                        outedges.append(noramp_edge)
                    if edge is None:
                        continue

                    plungelen = abs(edge.start_point[2] - edge.end_point[2])
                    projectionlen = plungelen * math.tan(
                        math.radians(rampangle)
                    )  # length of the forthcoming ramp projected to XY plane
                    # Path.Log.debug(
                    #    "Found plunge move at X:{} Y:{} From Z:{} to Z{}, length of ramp: {}".format(
                    #        p0.x, p0.y, p0.z, p1.z, projectionlen
                    #    )
                    # )
                    if self.method == "RampMethod3":
                        projectionlen = projectionlen / 2

                    # next need to determine how many edges in the path after
                    # plunge are needed to cover the length:
                    covered = False
                    coveredlen = 0
                    rampedges = []
                    i = edgei + 1
                    while not covered and i < len(edges):
                        candidate = edges[i]
                        if abs(candidate.start_point[2] - candidate.end_point[2]) > 1e-6 or (
                            not candidate.is_line and not candidate.is_arc
                        ):
                            # this edge is not an edge/arc in the XY plane; not qualified for ramping
                            break
                        # Path.Log.debug("Next edge length {}".format(candidate.Length))
                        rampedges.append(candidate)
                        coveredlen = coveredlen + candidate.xy_length

                        if coveredlen > projectionlen:
                            covered = True
                        i = i + 1
                    if len(rampedges) == 0:
                        Path.Log.warn("No suitable edges for ramping, plunge will remain as such")
                        outedges.append(edge)
                    else:
                        # Path.Log.debug("Doing ramp to edges: {}".format(rampedges))
                        if self.method == "RampMethod1":
                            outedges.extend(
                                self.createRampMethod1(
                                    rampedges, edge.start_point, projectionlen, rampangle
                                )
                            )
                        elif self.method == "RampMethod2":
                            outedges.extend(
                                self.createRampMethod2(
                                    rampedges, edge.start_point, projectionlen, rampangle
                                )
                            )
                        else:
                            # if the ramp cannot be covered with Method3, revert to Method1
                            # because Method1 support going back-and-forth and thus results in same path as Method3 when
                            # length of the ramp is smaller than needed for single ramp.
                            if not covered:
                                projectionlen = projectionlen * 2
                                outedges.extend(
                                    self.createRampMethod1(
                                        rampedges, edge.start_point, projectionlen, rampangle
                                    )
                                )
                            else:
                                outedges.extend(
                                    self.createRampMethod3(
                                        rampedges, edge.start_point, projectionlen, rampangle
                                    )
                                )
                else:
                    outedges.append(edge)
            else:
                outedges.append(edge)
        return outedges

    def generateHelix(self):
        edges = self.edges
        minZ = self.findMinZ(edges)
        Path.Log.debug("Minimum Z in this path is {}".format(minZ))
        outedges = []
        i = 0
        while i < len(edges):
            edge = edges[i]
            if edge.is_line or edge.is_arc:
                if edge.xy_length < 1e-6 and edge.end_point[2] < edge.start_point[2]:
                    noramp_edge, edge = self.processIgnoreAbove(edge)
                    if noramp_edge is not None:
                        outedges.append(noramp_edge)
                    if edge is None:
                        i = i + 1
                        continue

                    # next need to find a loop
                    loopFound = False
                    rampedges = []
                    j = i + 1
                    while j < len(edges) and not loopFound:
                        candidate = edges[j]
                        if abs(candidate.start_point[2] - candidate.end_point[2]) > 1e-6:
                            # this edge is not parallel to XY plane, not qualified for ramping.
                            # exit early, no loop found
                            break
                        if (
                            Path.Geom.isRoughly(edge.end_point[0], candidate.end_point[0])
                            and Path.Geom.isRoughly(edge.end_point[1], candidate.end_point[1])
                            and Path.Geom.isRoughly(edge.end_point[2], candidate.end_point[2])
                        ):
                            loopFound = True
                        rampedges.append(candidate)
                        j = j + 1
                    if not loopFound:
                        Path.Log.warn("No suitable helix found, leaving as a plunge")
                        outedges.append(edge)
                    else:
                        outedges.extend(self.createHelix(rampedges, edge.start_point[2]))
                        if not Path.Geom.isRoughly(edge.end_point[2], minZ):
                            # the edges covered by the helix not handled again,
                            # unless reached the bottom height
                            i = j - 1

                else:
                    outedges.append(edge)
            else:
                outedges.append(edge)
            i = i + 1
        return outedges

    """
    Edges, or parts of edges, above self.ignoreAbove should not be ramped.
    This method is a helper for splitting edges into a portion that should be
    ramped and a portion that should not be ramped.

    Returns (noramp_edge, ramp_edge). Either of these variables may be None
    """

    def processIgnoreAbove(self, edge):
        if not self.ignoreAboveEnabled:
            return None, edge
        z0, z1 = edge.start_point[2], edge.end_point[2]
        if z0 > self.ignoreAbove.Value:
            if z1 > self.ignoreAbove.Value or Path.Geom.isRoughly(z1, self.ignoreAbove.Value):
                # Entire plunge is above ignoreAbove
                return edge, None
            elif not Path.Geom.isRoughly(z0, self.ignoreAbove.Value):
                # Split the edge into regions above and below
                return (
                    edge.clone(z0, self.ignoreAbove.Value),
                    edge.clone(self.ignoreAbove.Value, z1),
                )
        # Entire plunge is below ignoreAbove
        return None, edge

    def createHelix(self, rampedges, startZ):
        outedges = []
        ramplen = 0
        for redge in rampedges:
            ramplen = ramplen + redge.xy_length
        rampheight = abs(startZ - rampedges[-1].end_point[2])

        max_rise_over_run = 1 / math.tan(math.radians(self.angle))
        num_loops = math.ceil(rampheight / ramplen / max_rise_over_run)
        rampedges *= num_loops
        ramplen *= num_loops

        rampangle_rad = math.atan(ramplen / rampheight)
        curZ = startZ
        for i, redge in enumerate(rampedges):
            deltaZ = redge.xy_length / math.tan(rampangle_rad)
            # compute new z, or clamp to end segment to avoid rounding error
            newZ = curZ - deltaZ if i < len(rampedges) - 1 else rampedges[-1].end_point[2]
            outedges.append(redge.clone(curZ, newZ))
            curZ = newZ
        return outedges

    def findMinZ(self, edges):
        minZ = 99999999999
        for edge in edges:
            if edge.command.Name in Path.Geom.CmdMoveAll and edge.end_point[2] < minZ:
                minZ = edge.end_point[2]
        return minZ

    def createRampMethod1(self, rampedges, p0, projectionlen, rampangle):
        """
        This method generates ramp with following pattern:
        1. Start from the original startpoint of the plunge
        2. Ramp down along the path that comes after the plunge
        3. When reaching the Z level of the original plunge, return back to the beginning
            by going the path backwards until the original plunge end point is reached
        4. Continue with the original path
        """
        ramp, reset = self._createRampMethod1(rampedges, p0, projectionlen, rampangle)
        return ramp + reset

    def _createRampMethod1(self, rampedges, p0, projectionlen, rampangle):
        """
        Helper method for generating ramps. Computes ramp method 1, but returns the result in pieces to allow for implementing the other ramp methods.
        Returns (ramp, reset)
        - ramp: array of commands ramping down
        - reset: array of commands returning from the bottom of the ramp to the bottom of the original plunge
        """
        ramp = []
        reset = []
        reversed_edges = [redge.clone(reverse=True) for redge in rampedges]
        rampremaining = projectionlen
        z = p0[2]  # start from the upper point of plunge
        goingForward = True
        i = 0  # current position = start of this edge. May be len(rampremaining) if going backwards
        while rampremaining > 0:
            redge = rampedges[i] if goingForward else reversed_edges[i - 1]

            # for i, redge in enumerate(rampedges):
            if redge.xy_length > rampremaining:
                # will reach end of ramp within this edge, needs to be split
                split_first, split_remaining = redge.split(rampremaining)
                ramp.append(split_first.clone(z_start=z))
                # now we have reached the end of the ramp. Go back to plunge position with constant Z
                # start that by going to the beginning of this splitEdge
                if goingForward:
                    reset.append(split_first.clone(reverse=True))
                else:
                    # if we were reversing, we continue to the same direction as the ramp
                    reset.append(split_remaining)
                    i = i - 1
                rampremaining = 0
                break
            else:
                deltaZ = redge.xy_length / math.tan(math.radians(rampangle))
                new_z = z - deltaZ
                ramp.append(redge.clone(z, new_z))
                z = new_z
                rampremaining = rampremaining - redge.xy_length
                i = i + 1 if goingForward else i - 1
                if i == 0:
                    goingForward = True
                if i == len(rampedges):
                    goingForward = False

        # now we need to return to original position.
        while i >= 1:
            reset.append(reversed_edges[i - 1])
            i = i - 1

        return ramp, reset

    def createRampMethod3(self, rampedges, p0, projectionlen, rampangle):
        """
        This method generates ramp with following pattern:
        1. Start from the original startpoint of the plunge
        2. Ramp down along the path that comes after the plunge until
           traveled half of the Z distance
        3. Change direction and ramp backwards to the original plunge end point
        4. Continue with the original path

        This path is computed using ramp method 1.
        """
        z_half = (p0[2] + rampedges[0].start_point[2]) / 2
        r1_rampedges = [redge.clone(z_half, z_half) for redge in rampedges]
        ramp, _ = self._createRampMethod1(r1_rampedges, p0, projectionlen, rampangle)
        ramp_back = [
            redge.clone(
                2 * z_half - redge.start_point[2], 2 * z_half - redge.end_point[2], reverse=True
            )
            for redge in ramp[::-1]
        ]
        return ramp + ramp_back

    def createRampMethod2(self, rampedges, p0, projectionlen, rampangle):
        """
        This method generates ramp with following pattern:
        1. Start from the original startpoint of the plunge
        2. Travel at start depth along the path, for a distance required for step 3
        3. Ramp backwards along the path at rampangle, arriving exactly at the bottom of the plunge
        4. Continue with the original path

        This path is computed using ramp method 1:
        1. Move all edges up to the start height
        2. Perform ramp method 1 from the bottom of the plunge *up* to the relocated path
        3. Reverse the resulting path (both edge order and direction)
        """
        r1_rampedges = [redge.clone(p0[2], p0[2]) for redge in rampedges]
        r1_p0 = rampedges[0].start_point
        r1_rampangle = -rampangle
        r1_result = self.createRampMethod1(r1_rampedges, r1_p0, projectionlen, r1_rampangle)
        outedges = [redge.clone(reverse=True) for redge in r1_result[::-1]]
        return outedges

    def createCommands(self, obj, edges):
        commands = [edge.command for edge in edges]
        outCommands = []

        tc = PathDressup.toolController(obj.Base)
        horizFeed = tc.HorizFeed.Value
        vertFeed = tc.VertFeed.Value
        horizRapid = tc.HorizRapid.Value
        vertRapid = tc.VertRapid.Value

        if obj.RampFeedRate == "Horizontal Feed Rate":
            rampFeed = horizFeed
        elif obj.RampFeedRate == "Vertical Feed Rate":
            rampFeed = vertFeed
        elif obj.RampFeedRate == "Ramp Feed Rate":
            rampFeed = math.sqrt(pow(vertFeed, 2) + pow(horizFeed, 2))
        else:
            rampFeed = obj.CustomFeedRate.Value

        lastX = lastY = lastZ = 0
        for cmd in commands:
            params = cmd.Parameters
            x = params.get("X", lastX)
            y = params.get("Y", lastY)
            z = params.get("Z", lastZ)

            z = z and round(z, 8)

            if cmd.Name in ["G1", "G2", "G3", "G01", "G02", "G03"]:
                if lastZ != z:
                    if Path.Geom.isRoughly(x, lastX) and Path.Geom.isRoughly(y, lastY):
                        params["F"] = vertFeed
                    else:
                        params["F"] = rampFeed
                else:
                    params["F"] = horizFeed

            elif cmd.Name in ["G0", "G00"]:
                if lastZ != z:
                    params["F"] = vertRapid
                else:
                    params["F"] = horizRapid

            lastX, lastY, lastZ = x, y, z

            outCommands.append(Path.Command(cmd.Name, params))

        return Path.Path(outCommands)


class ViewProviderDressup:
    def __init__(self, vobj):
        self.obj = vobj.Object

    def attach(self, vobj):
        self.obj = vobj.Object

    def claimChildren(self):
        if hasattr(self.obj.Base, "InList"):
            for i in self.obj.Base.InList:
                if hasattr(i, "Group"):
                    group = i.Group
                    for g in group:
                        if g.Name == self.obj.Base.Name:
                            group.remove(g)
                    i.Group = group
                    print(i.Group)
        # FreeCADGui.ActiveDocument.getObject(obj.Base.Name).Visibility = False
        return [self.obj.Base]

    def onDelete(self, arg1=None, arg2=None):
        """this makes sure that the base operation is added back to the project and visible"""
        Path.Log.debug("Deleting Dressup")
        if arg1.Object and arg1.Object.Base:
            FreeCADGui.ActiveDocument.getObject(arg1.Object.Base.Name).Visibility = True
            job = PathUtils.findParentJob(self.obj)
            if job:
                job.Proxy.addOperation(arg1.Object.Base, arg1.Object)
            arg1.Object.Base = None
        return True

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def getIcon(self):
        if getattr(PathDressup.baseOp(self.obj), "Active", True):
            return ":/icons/CAM_Dressup.svg"
        else:
            return ":/icons/CAM_OpActive.svg"


class CommandPathDressupRampEntry:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Dressup",
            "MenuText": QT_TRANSLATE_NOOP("CAM_DressupRampEntry", "Ramp Entry"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_DressupRampEntry",
                "Creates a ramp entry dress-up object from a selected toolpath",
            ),
        }

    def IsActive(self):
        op = PathDressup.selection()
        if op:
            return not PathDressup.hasEntryMethod(op)
        return False

    def Activated(self):

        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            Path.Log.error(translate("CAM_DressupRampEntry", "Select one toolpath object") + "\n")
            return
        baseObject = selection[0]
        if not baseObject.isDerivedFrom("Path::Feature"):
            Path.Log.error(
                translate("CAM_DressupRampEntry", "The selected object is not a toolpath") + "\n"
            )
            return
        if baseObject.isDerivedFrom("Path::FeatureCompoundPython"):
            Path.Log.error(translate("CAM_DressupRampEntry", "Select a Profile object"))
            return

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction("Create RampEntry Dress-up")
        FreeCADGui.addModule("Path.Dressup.Gui.RampEntry")
        FreeCADGui.addModule("PathScripts.PathUtils")
        FreeCADGui.doCommand(
            'obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", "RampEntryDressup")'
        )
        FreeCADGui.doCommand("dbo = Path.Dressup.Gui.RampEntry.ObjectDressup(obj)")
        FreeCADGui.doCommand("base = FreeCAD.ActiveDocument." + selection[0].Name)
        FreeCADGui.doCommand("job = PathScripts.PathUtils.findParentJob(base)")
        FreeCADGui.doCommand("obj.Base = base")
        FreeCADGui.doCommand("job.Proxy.addOperation(obj, base)")
        FreeCADGui.doCommand(
            "obj.ViewObject.Proxy = Path.Dressup.Gui.RampEntry.ViewProviderDressup(obj.ViewObject)"
        )
        FreeCADGui.doCommand("Gui.ActiveDocument.getObject(base.Name).Visibility = False")
        FreeCADGui.doCommand("dbo.setup(obj)")
        # FreeCAD.ActiveDocument.commitTransaction()  # Final `commitTransaction()` called via TaskPanel.accept()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_DressupRampEntry", CommandPathDressupRampEntry())

Path.Log.notice("Loading CAM_DressupRampEntry… done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Block objects."""
## @package make_block
# \ingroup draftmake
# \brief Provides functions to create Block objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.gui_utils as gui_utils

from draftobjects.block import Block

if App.GuiUp:
    from draftviewproviders.view_base import ViewProviderDraftPart


def make_block(objectslist):
    """make_block(objectslist)

    Creates a Draft Block from the given objects.

    Parameters
    ----------
    objectlist : list
        Major radius of the ellipse.

    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return
    obj = App.ActiveDocument.addObject("Part::FeaturePython", "Block")
    obj.addExtension("Part::AttachExtensionPython")
    Block(obj)
    obj.Components = objectslist
    if App.GuiUp:
        ViewProviderDraftPart(obj.ViewObject)
        for o in objectslist:
            o.ViewObject.Visibility = False
        gui_utils.select(obj)
    return obj


makeBlock = make_block

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 FreeCAD Team                                       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENSE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the DocumentObjectEditorWidget."""

import unittest
from unittest.mock import MagicMock
import FreeCAD
from PySide import QtGui
from Path.Tool.docobject import DetachedDocumentObject
from Path.Tool.docobject.ui.docobject import DocumentObjectEditorWidget, _get_label_text
from Path.Tool.docobject.ui.property import (
    BasePropertyEditorWidget,
    QuantityPropertyEditorWidget,
    BoolPropertyEditorWidget,
    IntPropertyEditorWidget,
    EnumPropertyEditorWidget,
    LabelPropertyEditorWidget,
)


class TestDocumentObjectEditorWidget(unittest.TestCase):
    """Tests for DocumentObjectEditorWidget."""

    def test_populate_form(self):
        obj = DetachedDocumentObject()
        obj.addProperty("App::PropertyString", "Prop1", "Group1", "Doc1")
        obj.Prop1 = "Value1"
        obj.addProperty("App::PropertyInt", "Prop2", "Group1", "Doc2")
        obj.Prop2 = 123
        obj.addProperty("App::PropertyLength", "Prop3", "Group1", "Doc3")
        obj.Prop3 = FreeCAD.Units.Quantity(5.0, "mm")
        obj.addProperty("App::PropertyBool", "Prop4", "Group1", "Doc4")
        obj.Prop4 = False
        obj.addProperty("App::PropertyEnumeration", "Prop5", "Group1", "Doc5")
        obj.Prop5 = ["OptionA", "OptionB"]

        properties_to_show = ["Prop1", "Prop2", "Prop3", "Prop4", "Prop5", "NonExistent"]
        property_suffixes = {"Prop1": "Suffix1", "Prop3": "Len"}

        widget = DocumentObjectEditorWidget(
            obj=obj, properties_to_show=properties_to_show, property_suffixes=property_suffixes
        )

        # Verify the layout contains the correct number of rows (excluding non-existent)
        expected_row_count = len([p for p in properties_to_show if hasattr(obj, p)])
        self.assertEqual(widget._layout.rowCount(), expected_row_count)

        # Verify labels and widgets are added correctly and are of the expected types
        prop_names_in_layout = []
        for i in range(widget._layout.rowCount()):
            label_item = widget._layout.itemAt(i, QtGui.QFormLayout.LabelRole)
            field_item = widget._layout.itemAt(i, QtGui.QFormLayout.FieldRole)

            self.assertIsNotNone(label_item)
            self.assertIsNotNone(field_item)

            label_widget = label_item.widget()
            field_widget = field_item.widget()

            self.assertIsInstance(label_widget, QtGui.QLabel)
            self.assertIsInstance(
                field_widget, BasePropertyEditorWidget
            )  # Check against base class

            # Determine the property name from the label text (reverse of _get_label_text)
            # This is a bit fragile, but necessary without storing prop_name in the label widget
            label_text = label_widget.text()
            prop_name = None
            for original_prop_name in properties_to_show:
                expected_label = _get_label_text(original_prop_name)
                suffix = property_suffixes.get(original_prop_name)
                if suffix:
                    expected_label = f"{expected_label} ({suffix}):"
                else:
                    expected_label = f"{expected_label}:"
                if label_text == expected_label:
                    prop_name = original_prop_name
                    break

            self.assertIsNotNone(
                prop_name, f"Could not determine property name for label: {label_text}"
            )
            prop_names_in_layout.append(prop_name)

            # Verify widget type based on property type
            if prop_name == "Prop1":
                self.assertIsInstance(field_widget, LabelPropertyEditorWidget)
                self.assertEqual(label_widget.text(), "Prop1 (Suffix1):")
            elif prop_name == "Prop2":
                self.assertIsInstance(field_widget, IntPropertyEditorWidget)
                self.assertEqual(label_widget.text(), "Prop2:")
            elif prop_name == "Prop3":
                self.assertIsInstance(field_widget, QuantityPropertyEditorWidget)
                self.assertEqual(label_widget.text(), "Prop3 (Len):")
            elif prop_name == "Prop4":
                self.assertIsInstance(field_widget, BoolPropertyEditorWidget)
                self.assertEqual(label_widget.text(), "Prop4:")
            elif prop_name == "Prop5":
                self.assertIsInstance(field_widget, EnumPropertyEditorWidget)
                self.assertEqual(label_widget.text(), "Prop5:")

        # Verify property editors are stored
        self.assertEqual(len(widget._property_editors), expected_row_count)
        for prop_name in prop_names_in_layout:
            self.assertIn(prop_name, widget._property_editors)
            self.assertIsInstance(widget._property_editors[prop_name], BasePropertyEditorWidget)

    def test_set_object(self):
        obj1 = DetachedDocumentObject()
        obj1.addProperty("App::PropertyString", "PropA", "GroupA", "DocA")
        obj1.PropA = "ValueA"

        obj2 = DetachedDocumentObject()
        obj2.addProperty("App::PropertyString", "PropA", "GroupA", "DocA")
        obj2.PropA = "ValueB"

        properties_to_show = ["PropA"]
        widget = DocumentObjectEditorWidget(obj=obj1, properties_to_show=properties_to_show)

        # Get the initial editor widget instance
        initial_editor = widget._property_editors["PropA"]
        self.assertIsInstance(initial_editor, BasePropertyEditorWidget)

        # Set a new object
        widget.setObject(obj2)

        # Verify that the editor widget instance is the same
        self.assertEqual(widget._property_editors["PropA"], initial_editor)
        # Verify that attachTo was called on the existing editor widget
        # This requires the real attachTo method to be implemented correctly
        # and the editor widget to update its internal object reference.
        # We can't easily assert the internal state change without mocks,
        # but we can trust the implementation of attachTo in PropertyEditorWidget.
        # We can verify updateWidget was called.
        # Note: This test relies on the side effect of attachTo calling updateWidget
        # in the real implementation.
        # We can't directly assert method calls without mocks, so we'll rely on
        # the fact that setting the object and then updating the UI should
        # reflect the new object's values if attachTo worked.
        widget.updateUI()
        # Check if the child widget's display reflects obj2's value
        # This requires accessing the child widget's internal editor widget
        # which might be fragile. A better approach is to trust the unit tests
        # for the individual PropertyEditorWidgets and focus on the
        # DocumentObjectEditorWidget's logic of calling attachTo and updateUI.

    def test_set_properties_to_show(self):
        obj = DetachedDocumentObject()
        obj.addProperty("App::PropertyString", "Prop1", "Group1", "Doc1")
        obj.Prop1 = "Value1"
        obj.addProperty("App::PropertyInt", "Prop2", "Group1", "Doc2")
        obj.Prop2 = 123

        widget = DocumentObjectEditorWidget(obj=obj, properties_to_show=["Prop1"])

        # Store the initial editor instance for Prop1
        initial_prop1_editor = widget._property_editors["Prop1"]

        # Set new properties to show
        new_properties_to_show = ["Prop1", "Prop2"]
        new_suffixes = {"Prop2": "Suffix2"}
        widget.setPropertiesToShow(new_properties_to_show, new_suffixes)

        # Verify that the form was repopulated with the new properties
        self.assertEqual(widget._layout.rowCount(), len(new_properties_to_show))

        # Verify property editors are updated and new ones created
        self.assertEqual(len(widget._property_editors), len(new_properties_to_show))
        self.assertIn("Prop1", widget._property_editors)
        self.assertIn("Prop2", widget._property_editors)

        # Verify that the editor for Prop1 is a *new* instance after repopulation
        self.assertIsNot(widget._property_editors["Prop1"], initial_prop1_editor)
        self.assertIsInstance(widget._property_editors["Prop2"], IntPropertyEditorWidget)

        # Verify labels including suffixes
        prop_names_in_layout = []
        for i in range(widget._layout.rowCount()):
            label_item = widget._layout.itemAt(i, QtGui.QFormLayout.LabelRole)
            label_widget = label_item.widget()
            label_text = label_widget.text()

            prop_name = None
            for original_prop_name in new_properties_to_show:
                expected_label = _get_label_text(original_prop_name)
                suffix = new_suffixes.get(original_prop_name)
                if suffix:
                    expected_label = f"{expected_label} ({suffix}):"
                else:
                    expected_label = f"{expected_label}:"
                if label_text == expected_label:
                    prop_name = original_prop_name
                    break
            prop_names_in_layout.append(prop_name)

        self.assertIn("Prop1", prop_names_in_layout)
        self.assertIn("Prop2", prop_names_in_layout)
        self.assertEqual(
            widget._layout.itemAt(0, QtGui.QFormLayout.LabelRole).widget().text(), "Prop1:"
        )
        self.assertEqual(
            widget._layout.itemAt(1, QtGui.QFormLayout.LabelRole).widget().text(),
            "Prop2 (Suffix2):",
        )

    def test_property_changed_signal(self):
        obj = DetachedDocumentObject()
        obj.addProperty("App::PropertyString", "Prop1", "Group1", "Doc1")
        obj.Prop1 = "Value1"

        widget = DocumentObjectEditorWidget(obj=obj, properties_to_show=["Prop1"])

        # Connect to the widget's propertyChanged signal
        mock_slot = MagicMock()
        widget.propertyChanged.connect(mock_slot)

        # Get the real child editor widget
        child_editor = widget._property_editors["Prop1"]
        self.assertIsInstance(child_editor, BasePropertyEditorWidget)

        # Emit the signal from the real child editor
        child_editor.propertyChanged.emit()

        # Verify that the widget's signal was emitted
        mock_slot.assert_called_once()

    def test_update_ui(self):
        obj = DetachedDocumentObject()
        obj.addProperty("App::PropertyString", "Prop1", "Group1", "Doc1")
        obj.Prop1 = "Value1"
        obj.addProperty("App::PropertyInt", "Prop2", "Group1", "Doc2")
        obj.Prop2 = 123

        properties_to_show = ["Prop1", "Prop2"]
        widget = DocumentObjectEditorWidget(obj=obj, properties_to_show=properties_to_show)

        # Get the real child editor widgets
        editor1 = widget._property_editors["Prop1"]
        editor2 = widget._property_editors["Prop2"]

        # Mock their updateWidget methods to check if they are called
        editor1.updateWidget = MagicMock()
        editor2.updateWidget = MagicMock()

        # Call updateUI
        widget.updateUI()

        # Verify that updateWidget was called on all child editors
        editor1.updateWidget.assert_called_once()
        editor2.updateWidget.assert_called_once()

    def test_update_object(self):
        obj = DetachedDocumentObject()
        obj.addProperty("App::PropertyString", "Prop1", "Group1", "Doc1")
        obj.Prop1 = "Value1"
        obj.addProperty("App::PropertyInt", "Prop2", "Group1", "Doc2")
        obj.Prop2 = 123

        properties_to_show = ["Prop1", "Prop2"]
        widget = DocumentObjectEditorWidget(obj=obj, properties_to_show=properties_to_show)

        # Get the real child editor widgets
        editor1 = widget._property_editors["Prop1"]
        editor2 = widget._property_editors["Prop2"]

        # Mock their updateProperty methods to check if they are called
        editor1.updateProperty = MagicMock()
        editor2.updateProperty = MagicMock()

        # Call updateObject
        widget.updateObject()

        # Verify that updateProperty was called on all child editors
        editor1.updateProperty.assert_called_once()
        editor2.updateProperty.assert_called_once()


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import json
import pathlib
from typing import Optional, Union, Sequence
import Path
from Path import Preferences
from Path.Preferences import addToolPreferenceObserver
from .assets import AssetManager, AssetUri, Asset, FileStore
from .toolbit.migration import ParameterAccessor, migrate_parameters

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


def ensure_library_assets_initialized(asset_manager: AssetManager, store_name: str = "local"):
    """
    Ensures the given store is initialized with built-in library
    if it is currently empty.
    """
    builtin_library_path = Preferences.getBuiltinLibraryPath()

    if asset_manager.is_empty("toolbitlibrary", store=store_name):
        for path in builtin_library_path.glob("*.fctl"):
            asset_manager.add_file("toolbitlibrary", path)


def ensure_toolbits_have_shape_type(asset_manager: AssetManager, store_name: str = "local"):
    from .shape import ToolBitShape

    toolbit_uris = asset_manager.list_assets(
        asset_type="toolbit",
        store=store_name,
    )

    for uri in toolbit_uris:
        data = asset_manager.get_raw(uri, store=store_name)
        attrs = json.loads(data)
        changed = False

        # --- Step 1: Ensure shape-type exists (migrate if needed) ---
        if "shape-type" not in attrs:
            shape_id = pathlib.Path(
                str(attrs.get("shape", ""))
            ).stem  # backward compatibility. used to be a filename
            if not shape_id:
                Path.Log.error(f"ToolBit {uri} missing shape ID")
                continue

            try:
                shape_class = ToolBitShape.get_shape_class_from_id(shape_id)
            except Exception as e:
                Path.Log.error(f"Failed to load toolbit {uri}: {e}. Skipping")
                continue
            if not shape_class:
                Path.Log.error(f"Toolbit {uri} has no shape-type attribute, and failed to infer it")
                continue
            attrs["shape-type"] = shape_class.name
            Path.Log.info(
                f"Migrating toolbit {uri}: Adding shape-type attribute '{shape_class.name}'"
            )
            changed = True

        # --- Step 2: Migrate legacy parameters (now that shape-type is set) ---
        if "parameter" in attrs and isinstance(attrs["parameter"], dict):
            if migrate_parameters(ParameterAccessor(attrs)):
                changed = True

        # --- Step 3: Write changes if any occurred ---
        if changed:
            data = json.dumps(attrs, sort_keys=True, indent=2).encode("utf-8")
            asset_manager.add_raw("toolbit", uri.asset_id, data, store=store_name)


def ensure_toolbit_assets_initialized(asset_manager: AssetManager, store_name: str = "local"):
    """
    Ensures the given store is initialized with built-in bits
    if it is currently empty.
    """
    builtin_toolbit_path = Preferences.getBuiltinToolBitPath()

    if asset_manager.is_empty("toolbit", store=store_name):
        for path in builtin_toolbit_path.glob("*.fctb"):
            asset_manager.add_file("toolbit", path)

    ensure_toolbits_have_shape_type(asset_manager, store_name)


def ensure_toolbitshape_assets_present(asset_manager: AssetManager, store_name: str = "local"):
    """
    Ensures the given store is initialized with built-in shapes
    if it is currently empty. This copies all built-in shapes,
    which is generally not recommended, but is useful for
    testing.

    In practice, the built-in tools don't need to be copied,
    because the CamAssetManager will automatically fall back to
    fetching them from the builtin store if they are not
    present in the local store (=the user's Shape directory).
    """
    builtin_shape_path = Preferences.getBuiltinShapePath()

    if asset_manager.is_empty("toolbitshape", store=store_name):
        for path in builtin_shape_path.glob("*.fcstd"):
            uri = AssetUri.build(
                asset_type="toolbitshape",
                asset_id=path.stem,
            )
            if not asset_manager.exists(uri, store=store_name):
                asset_manager.add_file("toolbitshape", path)

        for path in builtin_shape_path.glob("*.svg"):
            uri = AssetUri.build(
                asset_type="toolbitshapesvg",
                asset_id=path.stem + ".svg",
            )
            if not asset_manager.exists(uri, store=store_name):
                asset_manager.add_file("toolbitshapesvg", path, asset_id=path.stem + ".svg")

        for path in builtin_shape_path.glob("*.png"):
            uri = AssetUri.build(
                asset_type="toolbitshapepng",
                asset_id=path.stem + ".png",
            )
            if not asset_manager.exists(uri, store=store_name):
                asset_manager.add_file("toolbitshapepng", path, asset_id=path.stem + ".png")


def ensure_toolbitshape_assets_initialized(asset_manager: AssetManager, store_name: str = "local"):
    """
    Ensures the toolbitshape directory structure exists without adding any files.
    """
    from pathlib import Path

    # Get the shape directory path and ensure it exists
    shape_path = Preferences.getAssetPath() / "Tools" / "Shape"
    shape_path.mkdir(parents=True, exist_ok=True)


def ensure_assets_initialized(asset_manager: AssetManager, store="local"):
    """
    Ensures the given store is initialized with built-in assets.
    """
    ensure_library_assets_initialized(asset_manager, store)
    ensure_toolbit_assets_initialized(asset_manager, store)
    ensure_toolbitshape_assets_initialized(asset_manager, store)


def _on_asset_path_changed(group, key, value):
    Path.Log.info(f"CAM asset directory changed in preferences: {group} {key} {value}")
    user_asset_store.set_dir(value)
    ensure_assets_initialized(cam_assets)


# Set up the local CAM asset storage.
asset_mapping = {
    "toolbitlibrary": "Tools/Library/{asset_id}.fctl",
    "toolbit": "Tools/Bit/{asset_id}.fctb",
    "toolbitshape": "Tools/Shape/{asset_id}.fcstd",
    "toolbitshapesvg": "Tools/Shape/{asset_id}",  # Asset ID has ".svg" included
    "toolbitshapepng": "Tools/Shape/{asset_id}",  # Asset ID has ".png" included
    "machine": "Machine/{asset_id}.fcm",
}

# Separate mapping for builtin assets (maintains original structure)
builtin_asset_mapping = {
    "toolbitlibrary": "Library/{asset_id}.fctl",
    "toolbit": "Bit/{asset_id}.fctb",
    "toolbitshape": "Shape/{asset_id}.fcstd",
    "toolbitshapesvg": "Shape/{asset_id}",  # Asset ID has ".svg" included
    "toolbitshapepng": "Shape/{asset_id}",  # Asset ID has ".png" included
    "machine": "Machine/{asset_id}.fcm",
}

user_asset_store = FileStore(
    name="local",
    base_dir=Preferences.getAssetPath(),
    mapping=asset_mapping,
)

builtin_asset_store = FileStore(
    name="builtin",
    base_dir=Preferences.getBuiltinAssetPath(),
    mapping=builtin_asset_mapping,
)


class CamAssetManager(AssetManager):
    """
    Custom CAM Asset Manager that extends the base AssetManager, such
    that the get methods return fallbacks: if the asset is not present
    in the "local" store, then it falls back to the builtin-asset store.
    """

    def __init__(self):
        super().__init__()
        self.register_store(user_asset_store)
        self.register_store(builtin_asset_store)

    def setup(self):
        try:
            ensure_assets_initialized(cam_assets)
        except Exception as e:
            Path.Log.error(f"Failed to initialize CAM assets in {user_asset_store._base_dir}: {e}")
        else:
            Path.Log.debug(f"Using CAM assets in {user_asset_store._base_dir}")

    def get(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = ("local", "builtin"),
        depth: Optional[int] = None,
    ) -> Asset:
        """
        Gets an asset from the "local" store, falling back to the "builtin"
        store if not found locally.
        """
        return super().get(uri, store=store, depth=depth)

    def get_or_none(
        self,
        uri: Union[AssetUri, str],
        store: Union[str, Sequence[str]] = ("local", "builtin"),
        depth: Optional[int] = None,
    ) -> Optional[Asset]:
        """
        Gets an asset from the "local" store, falling back to the "builtin"
        store if not found locally.
        """
        return super().get_or_none(uri, store=store, depth=depth)


# Set up the CAM asset manager.
Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())
cam_assets = CamAssetManager()
addToolPreferenceObserver(_on_asset_path_changed)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Project"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchProject
#  \ingroup ARCH
#  \brief The Project object and tools
#
#  This module provides tools to build Project objects.

"""This module provides tools to build Project objects.  Project objects are
objects specifically for better IFC compatibility, allowing the user to tweak
certain IFC relevant values.
"""

import FreeCAD
import ArchIFC
import ArchIFCView

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:

    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt


class _Project(ArchIFC.IfcContext):
    """The project object.

    Takes a <Part::FeaturePython>, and turns it into a Project. Then takes a
    list of Arch sites to own as its children.

    Parameters
    ----------
    obj: <App::DocumentObjectGroupPython> or <App::FeaturePython>
        The object to turn into a Project.
    """

    def __init__(self, obj):
        obj.Proxy = self
        self.Type = "Project"
        self.setProperties(obj)
        obj.IfcType = "Project"

    def setProperties(self, obj):
        """Give the object properties unique to projects.

        Add the IFC context properties, and the group extension if it does not
        already exist.
        """

        ArchIFC.IfcContext.setProperties(self, obj)
        pl = obj.PropertiesList
        if not hasattr(obj, "Group"):
            obj.addExtension("App::GroupExtensionPython")

    def onDocumentRestored(self, obj):
        """Method run when the document is restored. Re-add the properties."""
        self.setProperties(obj)

    def dumps(self):

        return None

    def loads(self, state):

        self.Type = "Project"

    def addObject(self, obj, child):
        "Adds an object to the group of this BuildingPart"

        if not child in obj.Group:
            g = obj.Group
            g.append(child)
            obj.Group = g


class _ViewProviderProject(ArchIFCView.IfcContextView):
    """A View Provider for the project object.

    Parameters
    ----------
    vobj: <Gui.ViewProviderDocumentObject>
        The view provider to turn into a project view provider.
    """

    def __init__(self, vobj):
        vobj.Proxy = self
        vobj.addExtension("Gui::ViewProviderGroupExtensionPython")

    def getIcon(self):
        """Return the path to the appropriate icon.

        Returns
        -------
        str
            Path to the appropriate icon .svg file.
        """

        import Arch_rc

        return ":/icons/Arch_Project_Tree.svg"

    def removeDisplaymodeChildNodes(self, vobj):
        """Remove all child nodes from the 4 default display modes.

        This avoids 'ghosts' of the objects in the Group property.
        See:
        ArchSite.py
        https://forum.freecad.org/viewtopic.php?f=10&t=74731
        """

        from pivy import coin
        from draftutils import gui_utils

        if not hasattr(self, "displaymodes_cleaned"):
            if vobj.RootNode.getNumChildren():
                main_switch = gui_utils.find_coin_node(
                    vobj.RootNode, coin.SoSwitch
                )  # The display mode switch.
                if (
                    main_switch is not None and main_switch.getNumChildren() == 4
                ):  # Check if all display modes are available.
                    for node in tuple(main_switch.getChildren()):
                        node.removeAllChildren()
                    self.displaymodes_cleaned = True

    def onChanged(self, vobj, prop):
        self.removeDisplaymodeChildNodes(vobj)



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

from urllib.parse import urlparse

import addonmanager_freecad_interface as fci


def migrate_proxy_settings_2025():
    """Migrate the proxy settings from the old format to the new format, established in 2025"""
    if fci.Preferences().get("proxy_settings_migrated_2025"):
        return
    old_proxy_no = fci.Preferences().get("NoProxyCheck")
    old_proxy_system = fci.Preferences().get("SystemProxyCheck")
    old_proxy_user = fci.Preferences().get("UserProxyCheck")
    old_proxy_url = fci.Preferences().get("ProxyUrl")

    parsed_url = urlparse(old_proxy_url)

    new_proxy_host = parsed_url.hostname if parsed_url.hostname else ""
    new_proxy_port = int(parsed_url.port) if parsed_url.port else 8080

    if old_proxy_system:
        new_proxy_type = "system"
    elif old_proxy_user:
        new_proxy_type = "custom"
    elif old_proxy_no:
        new_proxy_type = "none"
    else:
        new_proxy_type = "system"

    fci.Preferences().set("proxy_type", new_proxy_type)
    fci.Preferences().set("proxy_host", new_proxy_host)
    fci.Preferences().set("proxy_port", new_proxy_port)
    fci.Preferences().set("proxy_settings_migrated_2025", True)



# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import unittest

import FreeCAD

App = FreeCAD


class TestDatumPoint(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestDatumPoint")

    def testOriginDatumPoint(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.DatumPoint = self.Doc.addObject("PartDesign::Point", "DatumPoint")
        self.DatumPoint.AttachmentSupport = [(self.Doc.XY_Plane, "")]
        self.DatumPoint.MapMode = "ObjectOrigin"
        self.Body.addObject(self.DatumPoint)
        self.Doc.recompute()
        self.assertEqual(self.DatumPoint.AttachmentOffset.Base, App.Vector(0))

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestDatumPoint")
        # print ("omit closing document for debugging")


class TestDatumLine(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestDatumLine")

    def testXAxisDatumLine(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.DatumLine = self.Doc.addObject("PartDesign::Line", "DatumLine")
        self.DatumLine.AttachmentSupport = [(self.Doc.XY_Plane, "")]
        self.DatumLine.MapMode = "ObjectX"
        self.Body.addObject(self.DatumLine)
        self.Doc.recompute()
        self.assertNotIn("Invalid", self.DatumLine.State)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestDatumLine")
        # print ("omit closing document for debugging")


class TestDatumPlane(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestDatumPlane")

    def testXYDatumPlane(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.DatumPlane = self.Doc.addObject("PartDesign::Plane", "DatumPlane")
        self.DatumPlane.AttachmentSupport = [(self.Doc.XY_Plane, "")]
        self.DatumPlane.MapMode = "FlatFace"
        self.Body.addObject(self.DatumPlane)
        self.Doc.recompute()
        self.DatumPlaneNormal = self.DatumPlane.Shape.Surface.Axis
        self.assertEqual(abs(self.DatumPlaneNormal.dot(App.Vector(0, 0, 1))), 1)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestDatumPlane")
        # print ("omit closing document for debugging")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2024 FreeCAD Project Association                        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Fillet objects between two lines.

This creates a `Part::Part2DObjectPython`, and then assigns the Proxy class
`Fillet`, and the `ViewProviderFillet` for the view provider.
"""
## @package make_fillet
# \ingroup draftmake
# \brief Provides functions to create Fillet objects between two lines.

import lazy_loader.lazy_loader as lz

import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils
import draftobjects.fillet as fillet

from draftutils.messages import _err
from draftutils.translate import translate

if App.GuiUp:
    import draftviewproviders.view_fillet as view_fillet

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")
DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")

## \addtogroup draftmake
# @{


def _preprocess(objs, radius, chamfer):
    """Check the inputs and return the edges for the fillet and the objects to be deleted."""
    edges = []
    del_objs = []
    if objs[0].isDerivedFrom("Gui::SelectionObject"):
        for sel in objs:
            for sub in sel.SubElementNames if sel.SubElementNames else [""]:
                shape = sel.Object.getSubObject(sub)
                if shape.ShapeType == "Edge":
                    edges.append(shape)
                    if sel.Object not in del_objs:
                        del_objs.append(sel.Object)
    else:
        for obj in objs:
            if hasattr(obj, "Shape"):
                shape = obj.Shape
                del_objs.append(obj)
            else:
                shape = obj
            if hasattr(shape, "ShapeType") and shape.ShapeType in ("Wire", "Edge"):
                edges.append(shape.Edges[0])

    if len(edges) != 2:
        _err(translate("draft", "2 edges are needed"))
        return None, None

    edges = DraftGeomUtils.fillet(edges, radius, chamfer)
    if len(edges) < 3:
        _err(translate("draft", "Edges are not connected or radius is too large"))
        return None, None

    return edges, del_objs


def make_fillet(objs, radius=100, chamfer=False, delete=False):
    """Create a fillet between two edges.

    Parameters
    ----------
    objs: list
        A list of two objects or shapes of type wire (1st edge is used) or edge,
        or a 2 edge selection set:`FreeCADGui.Selection.getSelectionEx("", 0)`.

    radius: float, optional
        It defaults to 100. The curvature of the fillet.

    chamfer: bool, optional
        It defaults to `False`. If it is `True` it no longer produces
        a rounded fillet but a chamfer (straight edge)
        with the value of the `radius`.

    delete: bool, optional
        It defaults to `False`. If `True` the source objects are deleted.
        Ignored for shapes.

    Returns
    -------
    Part::Part2DObjectPython
        The object of Proxy type `'Fillet'`.
        It returns `None` if it fails producing the object.
    """

    edges, del_objs = _preprocess(objs, radius, chamfer)
    if edges is None:
        return

    try:
        wire = Part.Wire(edges)
    except Part.OCCError:
        return None

    doc = App.activeDocument()
    obj = doc.addObject("Part::Part2DObjectPython", "Fillet")
    fillet.Fillet(obj)
    obj.Shape = wire
    obj.Length = wire.Length
    obj.Start = wire.Vertexes[0].Point
    obj.End = wire.Vertexes[-1].Point
    obj.FilletRadius = radius

    if delete:
        for del_obj in del_objs:
            doc.removeObject(del_obj.Name)

    if App.GuiUp:
        view_fillet.ViewProviderFillet(obj.ViewObject)
        gui_utils.format_object(obj)
        gui_utils.select(obj)

    return obj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# add import/export types

FreeCAD.addExportType("Industry Foundation Classes (*.ifc)", "importers.exportIFC")
# FreeCAD.addImportType("Industry Foundation Classes (*.ifc)","importIFC")
FreeCAD.addImportType("Industry Foundation Classes (*.ifc)", "nativeifc.ifc_import")
FreeCAD.addExportType("Industry Foundation Classes - IFCJSON (*.ifcJSON)", "importers.exportIFC")
FreeCAD.addImportType("Wavefront OBJ - Arch module (*.obj *.OBJ)", "importers.importOBJ")
FreeCAD.addExportType("Wavefront OBJ - Arch module (*.obj)", "importers.importOBJ")
FreeCAD.addExportType("WebGL file (*.html)", "importers.importWebGL")
FreeCAD.addExportType("JavaScript Object Notation (*.json)", "importers.importJSON")
FreeCAD.addImportType("Collada (*.dae *.DAE)", "importers.importDAE")
FreeCAD.addExportType("Collada (*.dae)", "importers.importDAE")
FreeCAD.addImportType("3D Studio mesh (*.3ds *.3DS)", "importers.import3DS")
FreeCAD.addImportType("SweetHome3D (*.sh3d)", "importers.importSH3D")
FreeCAD.addImportType("Shapefile (*.shp *.SHP)", "importers.importSHP")

FreeCAD.__unit_test__ += ["TestArch"]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint electric charge view provider"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package view_constraint_electricchargedensity
#  \ingroup FEM
#  \brief view provider for the constraint electric charge density object

from femtaskpanels import task_constraint_electricchargedensity
from . import view_base_femconstraint


class VPConstraintElectricChargeDensity(view_base_femconstraint.VPBaseFemConstraint):

    def __init__(self, vobj):
        super().__init__(vobj)
        mat = vobj.ShapeAppearance[0]
        mat.DiffuseColor = (1.0, 0.0, 0.2, 0.0)
        vobj.ShapeAppearance = mat

    def setEdit(self, vobj, mode=0):
        return view_base_femconstraint.VPBaseFemConstraint.setEdit(
            self, vobj, mode, task_constraint_electricchargedensity._TaskPanel
        )

    def attach(self, vobj):
        super().attach(vobj)
        vobj.loadSymbol(self.resource_symbol_dir + "ConstraintElectricChargeDensity.iv")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the Draft Workbench, object modification tests."""

## @package test_modification
# \ingroup drafttests
# \brief Unit tests for the Draft Workbench, object modification tests.

## \addtogroup drafttests
# @{

import FreeCAD as App
import Part
import Draft
from FreeCAD import Vector
from drafttests import auxiliary as aux
from drafttests import test_base
from draftutils.messages import _msg


class DraftModification(test_base.DraftTestCaseDoc):
    """Test Draft modification tools."""

    def test_move(self):
        """Create a line and move it."""
        operation = "Draft Move"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 2, 0)
        b = Vector(2, 2, 0)
        _msg("  Line")
        _msg("  a={0}, b={1}".format(a, b))
        obj = Draft.make_line(a, b)
        self.doc.recompute()

        c = Vector(3, 1, 0)
        _msg("  Translation vector")
        _msg("  c={}".format(c))
        Draft.move(obj, c)
        self.doc.recompute()
        self.assertTrue(obj.Start.isEqual(Vector(3, 3, 0), 1e-6), "'{}' failed".format(operation))

    def test_copy(self):
        """Create a line, then copy and move it."""
        operation = "Draft Move with copy"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 3, 0)
        b = Vector(2, 3, 0)
        _msg("  Line")
        _msg("  a={0}, b={1}".format(a, b))
        line = Draft.make_line(a, b)

        c = Vector(2, 2, 0)
        _msg("  Translation vector (copy)")
        _msg("  c={}".format(c))
        obj = Draft.move(line, c, copy=True)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_rotate(self):
        """Create a line, then rotate it."""
        operation = "Draft Rotate"
        _msg("  Test '{}'".format(operation))
        a = Vector(1, 1, 0)
        b = Vector(3, 1, 0)
        _msg("  Line")
        _msg("  a={0}, b={1}".format(a, b))
        obj = Draft.make_line(a, b)
        self.doc.recompute()

        c = Vector(-1, 1, 0)
        rot = 90
        _msg("  Rotation")
        _msg("  angle={} degrees".format(rot))
        Draft.rotate(obj, rot)
        self.doc.recompute()
        self.assertTrue(obj.Start.isEqual(c, 1e-6), "'{}' failed".format(operation))

    def test_offset_open(self):
        """Create an open wire, then produce an offset copy."""
        operation = "Draft Offset"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 2, 0)
        b = Vector(2, 4, 0)
        c = Vector(5, 2, 0)
        _msg("  Open wire")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  c={0}".format(c))
        wire = Draft.make_wire([a, b, c])
        self.doc.recompute()

        offset = Vector(-1, 1, 0)
        _msg("  Offset")
        _msg("  vector={}".format(offset))
        obj = Draft.offset(wire, offset, copy=True)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_offset_closed_with_reversed_edge(self):
        """Create a closed wire with a reversed edge, then produce an offset copy."""
        # Regression test for:
        # https://github.com/FreeCAD/FreeCAD/pull/5496
        operation = "Draft Offset"
        _msg("  Test '{}'".format(operation))
        _msg("  Closed wire with reversed edge")
        a = Vector(0, 0, 0)
        b = Vector(10, 0, 0)
        c = Vector(10, 4, 0)
        d = Vector(0, 4, 0)
        edges = [Part.makeLine(a, b), Part.makeLine(b, c), Part.makeLine(c, d), Part.makeLine(a, d)]
        wire = Part.Wire(edges)
        obj = self.doc.addObject("Part::Feature")
        obj.Shape = wire

        offset = Vector(0, -1, 0)
        new = Draft.offset(obj, offset, copy=True)
        self.assertTrue(len(new.Points) == 4, "'{}' failed".format(operation))

    def test_offset_rectangle_with_face(self):
        """Create a rectangle with a face, then produce an offset copy."""
        # Regression test for:
        # https://github.com/FreeCAD/FreeCAD/pull/7670
        operation = "Draft Offset"
        _msg("  Test '{}'".format(operation))
        length = 10
        width = 4
        _msg("  Rectangle with face")
        _msg("  length={0}, width={1}".format(length, width))
        rect = Draft.make_rectangle(length, width)
        rect.MakeFace = True
        self.doc.recompute()

        offset = Vector(0, -1, 0)
        _msg("  Offset")
        _msg("  vector={}".format(offset))
        obj = Draft.offset(rect, offset, copy=True)
        self.doc.recompute()
        obj_is_ok = (
            obj.Shape.CenterOfGravity == Vector(5, 2, 0) and obj.Length == 12 and obj.Height == 6
        )
        self.assertTrue(obj_is_ok, "'{}' failed".format(operation))

    def test_trim(self):
        """Trim a line. NOT IMPLEMENTED."""
        operation = "Draft Trimex trim"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(3, 3, 0)

        _msg("  Line")
        _msg("  a={0}, b={1}".format(a, b))
        line = Draft.make_line(a, b)

        c = Vector(2, 2, 0)
        d = Vector(4, 2, 0)
        _msg("  Line 2")
        _msg("  c={0}, d={1}".format(c, d))
        line2 = Draft.make_line(c, d)
        self.doc.recompute()

        obj = aux.fake_function(line, line2)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_extend(self):
        """Extend a line. NOT IMPLEMENTED."""
        operation = "Draft Trimex extend"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(1, 1, 0)
        _msg("  Line")
        _msg("  a={0}, b={1}".format(a, b))
        line = Draft.make_line(a, b)

        c = Vector(2, 2, 0)
        d = Vector(4, 2, 0)
        _msg("  Line 2")
        _msg("  c={0}, d={1}".format(c, d))
        line2 = Draft.make_line(c, d)
        self.doc.recompute()

        obj = aux.fake_function(line, line2)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_join(self):
        """Join two lines into a single Draft Wire."""
        operation = "Draft Join"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        _msg("  Line 1")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  Line 2")
        _msg("  b={0}, c={1}".format(b, c))
        line_1 = Draft.make_line(a, b)
        line_2 = Draft.make_line(b, c)

        # obj = Draft.join_wires([line_1, line_2])  # Multiple wires
        obj = Draft.join_two_wires(line_1, line_2)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_split(self):
        """Split a Draft Wire into two Draft Wires."""
        operation = "Draft_Split"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        d = Vector(6, 4, 0)
        _msg("  Wire")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  c={0}, d={1}".format(c, d))
        wire = Draft.make_wire([a, b, c, d])
        self.doc.recompute()

        index = 1
        _msg("  Split at")
        _msg("  p={0}, index={1}".format(b, index))
        obj = Draft.split(wire, b, index)
        # TODO: split needs to be modified so that it returns True or False.
        # Then checking for Wire001 is not needed
        if self.doc.Wire001:
            obj = True
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_upgrade(self):
        """Upgrade two Lines into a closed Wire, then draftify it."""
        operation = "Draft Upgrade"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        _msg("  Line 1")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  Line 2")
        _msg("  b={0}, c={1}".format(b, c))
        shape_line_1 = Part.makeLine(a, b)
        shape_line_2 = Part.makeLine(b, c)
        line_1 = self.doc.addObject("Part::Feature")
        line_2 = self.doc.addObject("Part::Feature")
        line_1.Shape = shape_line_1
        line_2.Shape = shape_line_2
        self.doc.recompute()

        # upgrade to wire
        obj = Draft.upgrade([line_1, line_2], delete=True)
        self.doc.recompute()
        s = obj[0][0]
        _msg("  1: Result '{0}' ({1})".format(s.Shape.ShapeType, s.TypeId))
        self.assertTrue(bool(obj[0]), "'{}' failed".format(operation))

        # upgrade to closed wire
        obj2 = Draft.upgrade(obj[0], delete=True)
        self.doc.recompute()
        s2 = obj2[0][0]
        _msg("  2: Result '{0}' ({1})".format(s2.Shape.ShapeType, s2.TypeId))
        self.assertTrue(bool(obj2[0]), "'{}' failed".format(operation))

        # upgrade to face
        obj3 = Draft.upgrade(obj2[0], delete=True)
        self.doc.recompute()
        s3 = obj3[0][0]
        _msg("  3: Result '{0}' ({1})".format(s3.Shape.ShapeType, s3.TypeId))
        self.assertTrue(bool(obj3[0]), "'{}' failed".format(operation))

        # upgrade to Draft_Wire
        Draft.upgrade(obj3[0], delete=True)
        # when draftifying, upgrade doesn't return a new object
        wire = self.doc.ActiveObject
        wire.MakeFace = True  # make test independent of MakeFaceMode parameter
        self.doc.recompute()
        _msg("  4: Result '{0}' ({1})".format(wire.Proxy.Type, wire.TypeId))
        self.assertTrue(bool(wire), "'{}' failed".format(operation))

        # Draft_Wire with face cannot be upgraded
        obj4 = Draft.upgrade(wire, delete=True)
        self.doc.recompute()
        _msg("  The last object cannot be upgraded further")
        self.assertFalse(bool(obj4[0]), "'{}' failed".format(operation))

    def test_downgrade(self):
        """Downgrade a closed Draft Wire into three simple Part Edges."""
        operation = "Draft Downgrade"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        _msg("  Closed wire")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  c={0}, a={1}".format(c, a))
        wire = Draft.make_wire([a, b, c, a])
        wire.MakeFace = True  # make test independent of MakeFaceMode parameter
        self.doc.recompute()

        # downgrade to face
        obj = Draft.downgrade(wire, delete=True)
        self.doc.recompute()
        s = obj[0][0]
        _msg("  1: Result '{0}' ({1})".format(s.Shape.ShapeType, s.TypeId))
        self.assertTrue(bool(obj[0]), "'{}' failed".format(operation))

        # downgrade to wire
        obj2 = Draft.downgrade(obj[0], delete=True)
        self.doc.recompute()
        s2 = obj2[0][0]
        _msg("  2: Result '{0}' ({1})".format(s2.Shape.ShapeType, s2.TypeId))
        self.assertTrue(bool(obj2[0]), "'{}' failed".format(operation))

        # downgrade to edges
        obj3 = Draft.downgrade(obj2[0], delete=True)
        self.doc.recompute()
        s3 = obj3[0][0]
        _msg("  3: Result 3 x '{0}' ({1})".format(s3.Shape.ShapeType, s3.TypeId))
        self.assertTrue(len(obj3[0]) == 3, "'{}' failed".format(operation))

        # edges cannot be downgraded
        obj4 = Draft.downgrade(obj3[0], delete=True)
        self.doc.recompute()
        s4 = obj4[0]
        _msg("  4: Result '{}'".format(s4))
        _msg("  The last objects cannot be downgraded further")
        self.assertFalse(bool(obj4[0]), "'{}' failed".format(operation))

    def test_wire_to_bspline(self):
        """Convert a polyline to BSpline and back."""
        operation = "Draft WireToBSpline"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        _msg("  Wire")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  c={}".format(c))
        wire = Draft.make_wire([a, b, c])

        obj = Draft.make_bspline(wire.Points)
        self.doc.recompute()
        _msg("  1: Result '{0}' ({1})".format(obj.Proxy.Type, obj.TypeId))
        self.assertTrue(obj, "'{}' failed".format(operation))

        obj2 = Draft.make_wire(obj.Points)
        _msg("  2: Result '{0}' ({1})".format(obj2.Proxy.Type, obj2.TypeId))
        self.assertTrue(obj2, "'{}' failed".format(operation))

    def test_shape_2d_view(self):
        """Create a prism and then a 2D projection of it."""
        operation = "Draft Shape2DView"
        _msg("  Test '{}'".format(operation))
        prism = self.doc.addObject("Part::Prism")
        prism.Polygon = 5
        # Rotate the prism 45 degrees around the Y axis
        prism.Placement.Rotation.Axis = Vector(0, 1, 0)
        prism.Placement.Rotation.Angle = 45 * (3.14159 / 180)
        _msg("  Prism")
        _msg("  n_sides={}".format(prism.Polygon))
        _msg("  placement={}".format(prism.Placement))

        direction = Vector(0, 0, 1)
        _msg("  Projection 2D view")
        _msg("  direction={}".format(direction))
        obj = Draft.make_shape2dview(prism, direction)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_draft_to_sketch(self):
        """Convert a Draft object to a Sketch and back."""
        operation = "Draft Draft2Sketch"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        _msg("  Wire")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  c={}".format(c))
        wire = Draft.make_wire([a, b, c])
        self.doc.recompute()

        obj = Draft.make_sketch(wire, autoconstraints=True)
        self.doc.recompute()
        _msg("  1: Result '{0}' ({1})".format(obj.Shape.ShapeType, obj.TypeId))
        self.assertTrue(obj, "'{}' failed".format(operation))

        obj2 = Draft.draftify(obj, delete=False)
        self.doc.recompute()
        _msg("  2: Result '{0}' ({1})".format(obj2.Proxy.Type, obj2.TypeId))
        self.assertTrue(obj2, "'{}' failed".format(operation))

    def test_rectangular_array(self):
        """Create a rectangle, and a rectangular array."""
        operation = "Draft OrthoArray"
        _msg("  Test '{}'".format(operation))
        length = 4
        width = 2
        _msg("  Rectangle")
        _msg("  length={0}, width={1}".format(length, width))
        rect = Draft.make_rectangle(length, width)
        self.doc.recompute()

        dir_x = Vector(5, 0, 0)
        dir_y = Vector(0, 4, 0)
        dir_z = Vector(0, 0, 6)
        number_x = 3
        number_y = 4
        number_z = 6
        _msg("  Array")
        _msg("  direction_x={}".format(dir_x))
        _msg("  direction_y={}".format(dir_y))
        _msg("  direction_z={}".format(dir_z))
        _msg("  number_x={0}, number_y={1}, number_z={2}".format(number_x, number_y, number_z))
        obj = Draft.make_ortho_array(rect, dir_x, dir_y, dir_z, number_x, number_y, number_z)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_polar_array(self):
        """Create a rectangle, and a polar array."""
        operation = "Draft PolarArray"
        _msg("  Test '{}'".format(operation))
        length = 4
        width = 2
        _msg("  Rectangle")
        _msg("  length={0}, width={1}".format(length, width))
        rect = Draft.make_rectangle(length, width)
        self.doc.recompute()

        center = Vector(-4, 0, 0)
        angle = 180
        number = 5
        _msg("  Array")
        _msg("  number={0}, polar_angle={1}".format(number, angle))
        _msg("  center={}".format(center))
        obj = Draft.make_polar_array(rect, number, angle, center)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_circular_array(self):
        """Create a rectangle, and a circular array."""
        operation = "Draft CircularArray"
        _msg("  Test '{}'".format(operation))
        length = 4
        width = 2
        _msg("  Rectangle")
        _msg("  length={0}, width={1}".format(length, width))
        rect = Draft.make_rectangle(length, width)
        self.doc.recompute()

        rad_distance = 10
        tan_distance = 8
        axis = Vector(0, 0, 1)
        center = Vector(0, 0, 0)
        number = 3
        symmetry = 1
        _msg("  Array")
        _msg("  radial_distance={0}, " "tangential_distance={1}".format(rad_distance, tan_distance))
        _msg("  number={0}, symmetry={1}".format(number, symmetry))
        _msg("  axis={}".format(axis))
        _msg("  center={}".format(center))
        obj = Draft.make_circular_array(
            rect, rad_distance, tan_distance, number, symmetry, axis, center
        )
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_path_array(self):
        """Create a wire, a polygon, and a path array."""
        operation = "Draft PathArray"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        d = Vector(8, 4, 0)
        _msg("  Wire")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  c={0}, d={1}".format(c, d))
        wire = Draft.make_wire([a, b, c, d])

        n_faces = 3
        radius = 1
        _msg("  Polygon")
        _msg("  n_faces={0}, radius={1}".format(n_faces, radius))
        poly = Draft.make_polygon(n_faces, radius)

        number = 4
        translation = Vector(0, 1, 0)
        subelements = "Edge1"
        align = False
        _msg("  Path Array")
        _msg("  number={}, translation={}".format(number, translation))
        _msg("  subelements={}, align={}".format(subelements, align))
        obj = Draft.make_path_array(poly, wire, number, translation, subelements, align)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_point_array(self):
        """Create a polygon, various point, and a point array."""
        operation = "Draft PointArray"
        _msg("  Test '{}'".format(operation))
        a = Vector(0, 0, 0)
        b = Vector(2, 2, 0)
        c = Vector(2, 4, 0)
        d = Vector(8, 4, 0)
        _msg("  Points")
        _msg("  a={0}, b={1}".format(a, b))
        _msg("  c={0}, d={1}".format(c, d))
        points = [
            Draft.make_point(a),
            Draft.make_point(b),
            Draft.make_point(c),
            Draft.make_point(d),
        ]

        _msg("  Upgrade")
        add, delete = Draft.upgrade(points)
        compound = add[0]

        n_faces = 3
        radius = 1
        _msg("  Polygon")
        _msg("  n_faces={0}, radius={1}".format(n_faces, radius))
        poly = Draft.make_polygon(n_faces, radius)

        _msg("  Point Array")
        obj = Draft.make_point_array(poly, compound)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_clone(self):
        """Create a box, then create a clone of it.

        Test for a bug introduced by changes in attachment code.
        """
        operation = "Draft Clone"
        _msg("  Test '{}'".format(operation))
        box = self.doc.addObject("Part::Box")
        self.doc.recompute()
        _msg("  object: '{0}' ({1})".format(box.Shape.ShapeType, box.TypeId))

        obj = Draft.make_clone(box)
        _msg("  clone: '{0}' ({1})".format(obj.Proxy.Type, obj.TypeId))
        self.assertTrue(obj, "'{}' failed".format(operation))
        self.assertTrue(obj.hasExtension("Part::AttachExtension"), "'{}' failed".format(operation))

    def test_attached_clone_behavior(self):
        """Check if an attached clone behaves correctly.

        Test for https://github.com/FreeCAD/FreeCAD/issues/8771.
        """
        operation = "Check attached Draft Clone behavior"
        _msg("  Test '{}'".format(operation))

        box1 = self.doc.addObject("Part::Box")
        box1.Length = 10
        box2 = self.doc.addObject("Part::Box")
        self.doc.recompute()

        obj = Draft.make_clone(box1)
        obj.MapMode = "ObjectXY"
        obj.AttachmentSupport = [(box2, ("",))]
        self.doc.recompute()

        box1.Length = 1
        self.doc.recompute()

        self.assertTrue(obj.Shape.BoundBox.XLength == 1, "'{}' failed".format(operation))

    def test_draft_to_techdraw(self):
        """Create a solid, and then a DraftView on a TechDraw page."""
        operation = "TechDraw DraftView (relies on Draft code)"
        _msg("  Test '{}'".format(operation))
        prism = self.doc.addObject("Part::Prism")
        prism.Polygon = 5
        # Rotate the prism 45 degrees around the Y axis
        prism.Placement.Rotation.Axis = Vector(0, 1, 0)
        prism.Placement.Rotation.Angle = 45 * (3.14159 / 180)
        _msg("  Prism")
        _msg("  n_sides={}".format(prism.Polygon))
        _msg("  placement={}".format(prism.Placement))

        page = self.doc.addObject("TechDraw::DrawPage")
        _msg("  page={}".format(page.TypeId))
        template = self.doc.addObject("TechDraw::DrawSVGTemplate")
        template.Template = (
            App.getResourceDir() + "Mod/TechDraw/Templates/ISO/A3_Landscape_blank.svg"
        )
        page.Template = template
        _msg("  template={}".format(template.TypeId))
        view = self.doc.addObject("TechDraw::DrawViewDraft")
        view.Source = prism
        view.Direction = App.Vector(0, 0, 1)
        page.addView(view)
        _msg("  view={}".format(view.TypeId))
        self.assertTrue(view, "'{}' failed".format(operation))
        self.assertTrue(view in page.OutList, "'{}' failed".format(operation))

    def test_mirror(self):
        """Create a rectangle, then a mirrored shape."""
        operation = "Draft Mirror"
        _msg("  Test '{}'".format(operation))
        length = 4
        width = 2
        _msg("  Rectangle")
        _msg("  length={0}, width={1}".format(length, width))
        rect = Draft.make_rectangle(length, width)
        # self.doc.recompute()

        p1 = Vector(6, -2, 0)
        p2 = Vector(6, 2, 0)
        _msg("  Mirror axis")
        _msg("  p1={}".format(p1))
        _msg("  p2={}".format(p2))
        obj = Draft.mirror(rect, p1, p2)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_stretch(self):
        """Stretch a line. NOT IMPLEMENTED."""
        operation = "Draft Stretch"
        _msg("  Test '{}'".format(operation))
        _msg("  This test requires an object and a selection")
        a = Vector(0, 0, 0)
        b = Vector(1, 1, 0)
        _msg("  Line")
        _msg("  a={0}, b={1}".format(a, b))
        line = Draft.make_line(a, b)
        direction = Vector(4, 1, 0)

        obj = aux.fake_function(line, direction)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_scale_part_feature_arcs(self):
        """Create and scale a part feature (arcs)."""
        operation = "Draft Scale part feature (arcs)"
        _msg("  Test '{}'".format(operation))

        base = Vector(3.5, 2.5, 0.0)
        cen = Vector(2.0, 1.0, 0.0)  # center for scaling
        sca = Vector(2.0, 3.0, 1.0)
        ends = [
            Vector(0.0, 0.0, 0.0),
            Vector(4.0, 0.0, 0.0),
            Vector(4.0, 3.0, 0.0),
            Vector(0.0, 3.0, 0.0),
        ]
        mids = [
            Vector(2.0, -0.5, 0.0),
            Vector(4.5, 1.5, 0.0),
            Vector(2.0, 3.5, 0.0),
            Vector(-0.5, 1.5, 0.0),
        ]  # arc midpoints

        shp = Part.Shape(
            [
                Part.Arc(ends[0], mids[0], ends[1]),
                Part.Arc(ends[1], mids[1], ends[2]),
                Part.Arc(ends[2], mids[2], ends[3]),
                Part.Arc(ends[3], mids[3], ends[0]),
            ]
        )
        obj = self.doc.addObject("Part::Feature")
        obj.Shape = shp
        obj.Placement.Base = base
        self.doc.recompute()
        obj = Draft.scale(obj, sca, cen, False)
        self.doc.recompute()

        # check endpoints of arcs:
        newEnds = [
            Vector(5.0, 5.5, 0.0),
            Vector(13.0, 5.5, 0.0),
            Vector(13.0, 14.5, 0.0),
            Vector(5.0, 14.5, 0.0),
        ]
        vrts = obj.Shape.Vertexes
        for i in range(4):
            self.assertTrue(
                vrts[i].Point.isEqual(newEnds[i], 1e-6), "'{}' failed".format(operation)
            )
        # check midpoints of arcs:
        newMids = [
            Vector(9.0, 4.0, 0.0),
            Vector(14.0, 10.0, 0.0),
            Vector(9.0, 16.0, 0.0),
            Vector(4.0, 10.0, 0.0),
        ]
        for i in range(4):
            edge = obj.Shape.Edges[i]
            par = (edge.LastParameter - edge.FirstParameter) / 2.0
            self.assertTrue(
                edge.valueAt(par).isEqual(newMids[i], 1e-6), "'{}' failed".format(operation)
            )

    def test_scale_part_feature_lines(self):
        """Create and scale a part feature (lines)."""
        operation = "Draft Scale part feature (lines)"
        _msg("  Test '{}'".format(operation))

        base = Vector(3.5, 2.5, 0.0)
        cen = Vector(2.0, 1.0, 0.0)  # center for scaling
        sca = Vector(2.0, 3.0, 1.0)
        pts = [
            Vector(0.0, 0.0, 0.0),
            Vector(4.0, 0.0, 0.0),
            Vector(4.0, 3.0, 0.0),
            Vector(0.0, 3.0, 0.0),
        ]

        shp = Part.Shape(
            [
                Part.LineSegment(pts[0], pts[1]),
                Part.LineSegment(pts[1], pts[2]),
                Part.LineSegment(pts[2], pts[3]),
                Part.LineSegment(pts[3], pts[0]),
            ]
        )
        obj = self.doc.addObject("Part::Feature")
        obj.Shape = shp
        obj.Placement.Base = base
        self.doc.recompute()
        obj = Draft.scale(obj, sca, cen, False)
        self.doc.recompute()

        newPts = [
            Vector(5.0, 5.5, 0.0),
            Vector(13.0, 5.5, 0.0),
            Vector(13.0, 14.5, 0.0),
            Vector(5.0, 14.5, 0.0),
        ]
        vrts = obj.Shape.Vertexes
        for i in range(4):
            self.assertTrue(vrts[i].Point.isEqual(newPts[i], 1e-6), "'{}' failed".format(operation))

    def test_scale_rectangle(self):
        """Create and scale a rectangle."""
        operation = "Draft Scale rectangle"
        _msg("  Test '{}'".format(operation))

        base = Vector(3.5, 2.5, 0.0)
        cen = Vector(2.0, 1.0, 0.0)  # center for scaling
        sca = Vector(2.0, 3.0, 1.0)
        len = 4.0
        hgt = 3.0

        obj = Draft.make_rectangle(len, hgt)
        obj.Placement.Base = base
        self.doc.recompute()
        obj = Draft.scale(obj, sca, cen, False)
        self.doc.recompute()

        newBase = Vector(5.0, 5.5, 0.0)
        newLen = 8.0
        newHgt = 9.0
        self.assertTrue(obj.Placement.Base.isEqual(newBase, 1e-6), "'{}' failed".format(operation))
        self.assertAlmostEqual(obj.Length, newLen, delta=1e-6, msg="'{}' failed".format(operation))
        self.assertAlmostEqual(obj.Height, newHgt, delta=1e-6, msg="'{}' failed".format(operation))

    def test_scale_spline(self):
        """Create and scale a spline."""
        operation = "Draft Scale spline"
        _msg("  Test '{}'".format(operation))

        base = Vector(3.5, 2.5, 0.0)
        cen = Vector(2.0, 1.0, 0.0)  # center for scaling
        sca = Vector(2.0, 3.0, 1.0)
        pts = [Vector(0.0, 0.0, 0.0), Vector(2.0, 3.0, 0.0), Vector(4.0, 0.0, 0.0)]

        obj = Draft.make_bspline(pts, False)
        obj.Placement.Base = base
        self.doc.recompute()
        obj = Draft.scale(obj, sca, cen, False)
        self.doc.recompute()
        pla = obj.Placement

        newPts = [Vector(5.0, 5.5, 0.0), Vector(9.0, 14.5, 0.0), Vector(13.0, 5.5, 0.0)]
        for i in range(3):
            self.assertTrue(
                pla.multVec(obj.Points[i]).isEqual(newPts[i], 1e-6), "'{}' failed".format(operation)
            )

    def test_scale_wire(self):
        """Create and scale a wire."""
        operation = "Draft Scale wire"
        _msg("  Test '{}'".format(operation))

        base = Vector(3.5, 2.5, 0.0)
        cen = Vector(2.0, 1.0, 0.0)  # center for scaling
        sca = Vector(2.0, 3.0, 1.0)
        pts = [
            Vector(0.0, 0.0, 0.0),
            Vector(4.0, 0.0, 0.0),
            Vector(4.0, 3.0, 0.0),
            Vector(0.0, 3.0, 0.0),
        ]

        obj = Draft.make_wire(pts, True)
        obj.Placement.Base = base
        self.doc.recompute()
        obj = Draft.scale(obj, sca, cen, False)
        self.doc.recompute()

        newPts = [
            Vector(5.0, 5.5, 0.0),
            Vector(13.0, 5.5, 0.0),
            Vector(13.0, 14.5, 0.0),
            Vector(5.0, 14.5, 0.0),
        ]
        vrts = obj.Shape.Vertexes
        for i in range(4):
            self.assertTrue(vrts[i].Point.isEqual(newPts[i], 1e-6), "'{}' failed".format(operation))


## @}



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Material FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import unittest
from os.path import join

import FreeCAD

from .support_utils import fcc_print


class TestMaterialUnits(unittest.TestCase):
    fcc_print("import TestMaterialUnits")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars

        fcc_print(
            "\n{0}\n{1} run FEM TestMaterialUnits tests {2}\n{0}".format(
                100 * "*", 10 * "*", 57 * "*"
            )
        )

    # ********************************************************************************************
    def test_known_quantity_units(self):
        from materialtools.cardutils import get_known_material_quantity_parameter as knownquant

        known_quantity_parameter = knownquant()
        from materialtools.cardutils import check_parm_unit as checkparamunit

        for param in known_quantity_parameter:
            fcc_print(f"{param}")
            self.assertTrue(
                checkparamunit(param),
                "Unit of quantity material parameter {} "
                "is unknown to the FreeCAD unit system.".format(param),
            )

    # ********************************************************************************************
    def test_material_card_quantities(self):
        # test the value and unit of known quantity parameter
        # from solid build in material cards
        # keep in mind only if FreeCAD is installed all materials are copied
        # TODO Fluid materials (are they installed?)

        # get build in materials
        builtin_solid_mat_dir = join(
            FreeCAD.getResourceDir(), "Mod", "Material", "StandardMaterial"
        )
        fcc_print(f"{builtin_solid_mat_dir}")
        from materialtools.cardutils import add_cards_from_a_dir as addmats

        materials, cards, icons = addmats({}, {}, {}, builtin_solid_mat_dir, "")

        # get known material quantity parameter
        from materialtools.cardutils import get_known_material_quantity_parameter as knownquant

        known_quantities = knownquant()

        # check param, value pairs
        from materialtools.cardutils import check_value_unit as checkvalueunit

        for mat in materials:
            fcc_print(f"{mat}")
            for param, value in materials[mat].items():
                if param in known_quantities:
                    # fcc_print("    {} --> {}".format(param, value))
                    self.assertTrue(
                        checkvalueunit(param, value),
                        "Unit of quantity {} from material parameter {} is wrong.".format(
                            value, param
                        ),
                    )
